/*   ftps.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996,1997,1998,1999 by Christoph Schwarz
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include "tweak.h"
#include "globals.h"

void* alloc(void* buf, long size)
{ if (buf) return realloc(buf,size); else return malloc(size);
}

void quit(int reason)
{ close_server = 1;
  if (close_reason == CL_ERROR) close_reason = reason;
}

void signal_handler(int sig)
{ if (((sig_write==sig_size-1) && !sig_read) || (sig_write+1==sig_read))
    return;
  sig_array[sig_write++] = sig;
  if (sig_write >= sig_size) sig_write = 0;
}

void handle_signals(void)
{ while (sig_read != sig_write) switch (sig_array[sig_read])
  { case SIGHUP :
    case SIGINT :
    case SIGQUIT :
    case SIGTERM :
      quit(sig_array[sig_read] == SIGQUIT ? CL_KICK : CL_SHUTDOWN);
      if (sig_read+1 >= sig_size) sig_read = 0; else sig_read++;
      break;
    case SIGPIPE :
      quit(CL_PIPEERR);
      if (sig_read+1 >= sig_size) sig_read = 0; else sig_read++;
      break;
    case SIGALRM :
      quit(CL_TIMEOUT);
    default :
      if (sig_read+1 >= sig_size) sig_read = 0; else sig_read++;
  }
}

void get_group_list(void)
{ struct group_s* g;
  int             i;

  printf("GLR\n");
  if (!scanf("%i",&group_count)) return;
  getc(stdin);
  if (group_count > group_size)
  { group_size = group_count;
    group = (struct group_s*)alloc(group,group_size*sizeof(struct group_s));
    if (!group)
    { fprintf(stderr,"ftps(%i): not enough memory for group list.\n",getpid());
      exit(-1);
    }
  }
  printf("+\n");
  g = group;
  for (i=0; i<group_count; i++)
  { if (scanf("%s %i",&g->name,&g->gid) != 2)
    { fprintf(stderr,"ftps(%i): error reading group list.\n",getpid());
      exit(-1);
    }
    getc(stdin);
    g++;
  }
}

void get_user_list(void)
{ int            i;
  struct user_s* u;

  printf("ULR\n");
  if (!scanf("%i",&user_count)) return;
  getc(stdin);
  if (user_count > user_size)
  { user_size = user_count;
    user = (struct user_s*)alloc(user,user_size*sizeof(struct user_s));
    if (!user)
    { fprintf(stderr,"ftps(%i): not enough memory for user list.\n",getpid());
      quit(CL_PIPEERR);
      exit(-1);
    }
  }
  printf("+\n");
  u = user;
  for (i=0; i<user_count; i++)
  { if (scanf("%s %i",&u->name,&u->uid) != 2)
    { fprintf(stderr,"ftps(%i): error reading user list.\n",getpid());
      quit(CL_PIPEERR);
      exit(-1);
    }
    getc(stdin);
    u++;
  }
}

void get_cdpath(void)
{ long gid;
  long perm;
  int  size;
  long uid;

  printf("SCD\n");
  if (!scanf("%i",&size)) return;
  getc(stdin);
  if (!size) return;
  if (size > cdpath_size)
  { cdpath_size = size;
    cdpath = alloc(cdpath,size);
    if (!cdpath)
    { fprintf(stderr,"ftps(%i): not enough memory for cdpath.\n",getpid());
      quit(CL_PIPEERR);
      exit(-1);
    }
  }
  printf("+\n");
  size = 0;
  while (size < cdpath_size)
  { fgets(cdpath+size,cdpath_size-size,stdin);
    if ('\n' == (cdpath+size)[strlen(cdpath+size)-1]) (cdpath+size)[strlen(cdpath+size)-1] = '\0';
    size += (strlen(cdpath+size) + 8) & ~0x07;
    if (scanf("%li %li %li",&uid,&gid,&perm) != 3)
    { fprintf(stderr,"ftps(%i): error reading cdpath.\n",getpid());
      quit(CL_PIPEERR);
      exit(-1);
    }
    *(long*)(cdpath+size) = uid;
    *(long*)(cdpath+size+sizeof(long)) = gid;
    *(long*)(cdpath+size+(sizeof(long)*2)) = perm;
    size += sizeof(long) * 3;
    getc(stdin);
  }
}

void get_checks(char* cmd, struct check_s** check, int* check_count)
{ struct check_s* ch;
  int             i;

  printf("%s\n",cmd);
  if (!scanf("%i",check_count)) return;
  getc(stdin);
  if (!*check_count) return;
  *check = (struct check_s*)alloc((char*)*check,
    *check_count*sizeof(struct check_s));
  if (!*check)
  { fprintf(stderr,"ftps(%i): not enough memory for check.\n",getpid());
    quit(CL_PIPEERR);
    exit(-1);
  }
  printf("+\n");
  ch = *check;
  for (i=0; i<*check_count; i++, ch++)
  { if (scanf("%s %s %s",ch->extension,ch->name,ch->param) != 3)
    { fprintf(stderr,"ftps(%i): error reading check.\n",getpid());
      quit(CL_PIPEERR);
      exit(-1);
    }
    if (!strcmp(ch->param,"---")) *ch->param = '\0';
    getc(stdin);
  }
}

void get_check(void)
{ get_checks("SCH",&check,&check_count);
  get_checks("SDC",&dcheck,&dcheck_count);
}

void get_dshortcut(struct sshortcut_s** dshortcut2)
{ int n = 0;
  int count=0;
  struct sshortcut_s* dsh;

  printf("SDS\n");
  scanf("%i",&dshortcut_count);
  getc(stdin);
  if (!dshortcut_count) return;
  dsh=(struct sshortcut_s *)calloc(dshortcut_count,sizeof(struct sshortcut_s));
  if (!dsh)
  { printf("-\n");
    fprintf(stderr,"ftps(%i): malloc error (dshortcut)\n", getpid());
    exit(close_reason);
  }
  printf("+\n");
  n = 0;
  count=0;
  while (n < dshortcut_count)
  {
    if (!fgets(dsh[count].shortname,sizeof(dsh[count].shortname)+10,stdin))
    { fprintf(stderr,"ftps(%i): fgets error (dshortcut)\n", getpid());
      quit(CL_PIPEERR);
      exit(close_reason);
    }
    *(dsh[count].shortname + strlen(dsh[count].shortname)-1) = '\0';
    if (!fgets(dsh[count].longname,sizeof(dsh[count].longname)+1,stdin))
    { fprintf(stderr,"ftps(%i): fgets error (dshortcut)\n", getpid());
      quit(CL_PIPEERR);
      exit(close_reason);
    }
    *(dsh[count].longname+strlen(dsh[count].longname)-1) = '\0';
    if(dsh[count].shortname[0]!='\0')
      count++;
    n++;
  }
  dshortcut = (struct sshortcut_s* )calloc(count,sizeof(struct sshortcut_s));
  if (!dshortcut)
  { fprintf(stderr,"ftps(%i): malloc error (dshortcut)\n", getpid());
    exit(close_reason);
  }
  for(n=0;n<count;n++)
  { strcpy(dshortcut[n].shortname,dsh[n].shortname);
    strcpy(dshortcut[n].longname,dsh[n].longname);
  }
  if(dsh)
    free(dsh);
  dshortcut_size=sizeof(dshortcut);
}


void get_dupecheck(void)
{ int n = 0;

  printf("DUP\n");
  scanf("%i",&dupechk_size);
  getc(stdin);
  if (!dupechk_size) return;
  dupechk = malloc(dupechk_size+1);
  if (!dupechk)
  { printf("-\n");
    fprintf(stderr,"ftps(%i): malloc error (dupechk)\n",getpid());
    exit(close_reason);
  }
  printf("+\n");
  n = 0;
  while (n < dupechk_size)
  { if (!fgets(dupechk+n,dupechk_size-n+2,stdin))
    { fprintf(stderr,"ftps(%i): fgets error\n",getpid());
      quit(CL_PIPEERR);
      exit(close_reason);
    }
    n += strlen(dupechk+n) - 1;
    *(dupechk+n) = '\0';
    n++;
  }
  *(dupechk+dupechk_size) = 0;
}

void get_top(void)
{ int             i;

  printf("STO\n");
  for (i=0; i<10; i++)
  { if (scanf("%s %li %s %f %s %f %s %li",top[i].d_name,&top[i].d_amount,
      top[i].e_name,&top[i].e_amount,top[i].l_name,&top[i].l_amount,
      top[i].u_name,&top[i].u_amount) != 8)
    { fprintf(stderr,"ftps(%i): error reading top list.\n",getpid());
      quit(CL_PIPEERR);
      exit(-1);
    }
    getc(stdin);
    fgets(top[i].d_info,sizeof(top[i].d_info), stdin);
    fgets(top[i].e_info,sizeof(top[i].e_info), stdin);
    fgets(top[i].l_info,sizeof(top[i].l_info), stdin);
    fgets(top[i].u_info,sizeof(top[i].u_info), stdin);
  }
}

int cmd_abor();
int cmd_acct();
int cmd_allo();
int cmd_appe();
int cmd_cdup();
int cmd_cwd();
int cmd_dele();
int cmd_help();
int cmd_list();
int cmd_mdtm();
int cmd_mkd();
int cmd_mode();
int cmd_nlst();
int cmd_noop();
int cmd_pass();
int cmd_pasv();
int cmd_port();
int cmd_pwd();
int cmd_quit();
int cmd_rein();
int cmd_rest();
int cmd_retr();
int cmd_rmd();
int cmd_rnfr();
int cmd_rnto();
int cmd_site();
int cmd_size();
int cmd_smnt();
int cmd_stat();
int cmd_stor();
int cmd_stou();
int cmd_stru();
int cmd_syst();
int cmd_type();
int cmd_user();

#define FTPCMD_CNT 35

struct ftpcmd_s
{ char command[4];
  int  (*function)(char**,int,char*,int);
  int  must_be_loggedin;
  int  parse_command;
} ftpcmd[FTPCMD_CNT] =
{ { "ABOR", cmd_abor, 1, 0 }, { "ACCT", cmd_acct, 1, 0 }, { "ALLO", cmd_allo, 1, 0 },
  { "APPE", cmd_appe, 1, 0 }, { "CDUP", cmd_cdup, 1, 0 }, { "CWD",  cmd_cwd,  1, 0 },
  { "DELE", cmd_dele, 1, 0 }, { "HELP", cmd_help, 1, 0 }, { "LIST", cmd_list, 1, 1 },
  { "MDTM", cmd_mdtm, 1, 0 }, { "MKD",  cmd_mkd,  1, 0 }, { "MODE", cmd_mode, 0, 0 },
  { "NLST", cmd_nlst, 1, 1 }, { "NOOP", cmd_noop, 1, 0 }, { "PASS", cmd_pass, 1, 0 },
  { "PASV", cmd_pasv, 1, 0 }, { "PORT", cmd_port, 1, 0 }, { "PWD",  cmd_pwd,  1, 0 },
  { "QUIT", cmd_quit, 0, 0 }, { "REIN", cmd_rein, 1, 0 }, { "REST", cmd_rest, 1, 0 },
  { "RETR", cmd_retr, 1, 0 }, { "RMD",  cmd_rmd,  1, 0 }, { "RNFR", cmd_rnfr, 1, 0 },
  { "RNTO", cmd_rnto, 1, 0 }, { "SITE", cmd_site, 1, 1 }, { "SIZE", cmd_size, 1, 0 },
  { "SMNT", cmd_smnt, 1, 0 }, { "STAT", cmd_stat, 1, 0 }, { "STOR", cmd_stor, 1, 0 },
  { "STOU", cmd_stou, 1, 0 }, { "STRU", cmd_stru, 0, 0 }, { "SYST", cmd_syst, 0, 0 },
  { "TYPE", cmd_type, 0, 1 }, { "USER", cmd_user, 0, 0 }} ;

int dispatch_command(char* buffer, int bufferlen, char** p, int n)
{ int i = 0;
  int j;
  int k = FTPCMD_CNT - 1;
  int r;

  while (i <= k)
  { j = (i + k) >> 1;
    r = strncasecmp(*p,ftpcmd[j].command,4);
    if (r < 0) k = j - 1; else if (r > 0) i = j + 1; else
    { if (!logged_in && ftpcmd[j].must_be_loggedin)
      { retries++;
        if (login_retries && (retries >= login_retries))
        { quit(CL_NORMAL);
          client_log_start(LG_LOGIN,"LOGIN ");
          client_log_finish(LG_LOGIN,"FAILED: Not logged in yet");
          return send_reply(421,"Not logged in yet, connection closed.");
        }
        return send_reply(503,"Please login first.");
      }
      if ((n == 2) && ftpcmd[j].parse_command)
      { n--;
        if (parse_command(*(p+1),p,31,&n))
          return send_reply(500,"Too many parameters.");
      }
      return ftpcmd[j].function(p+1,n-1,buffer,bufferlen);
    }
  }
  return send_reply(500,"Unrecognised command.");
}

void get_string(char** buffer, char* str, int max)
{ fgets(str, max, stdin);
  if (*str)
  { *buffer = malloc(strlen(str)+1);
    if (!*buffer)
    { fprintf(stderr,"ftps(%i): out of memory\n",getpid());
      exit(close_reason);
    }
    if ('\n' == str[strlen(str)-1]) str[strlen(str)-1] = '\0';
    strcpy(*buffer,str);
  } else *buffer = NULL;
}

int get_message(char* cmd, char** message, int* msg_len, int* msg_size)
{ int n;
  int mlen = 0;
  int msize = 0;

  if (!msg_len) msg_len = &mlen;
  if (!msg_size) msg_size = &msize;
  printf("%s\n",cmd);
  scanf("%i",msg_len);
  getc(stdin);
  if (!*msg_len) return 1;
  if (*msg_len >= *msg_size)
  { *msg_size = *msg_len+1;
    *message = alloc(*message,*msg_size+2);
  }
  if (!*message)
  { printf("-\n");
    fprintf(stderr,"ftps(%i): alloc error\n",getpid());
    return 0;
  }
  printf("+\n");
  n = 0;
  while (n < *msg_len)
  { if (!fgets(*message+n,*msg_size-n+2,stdin))
    { fprintf(stderr,"ftps(%i): fgets error (get_message)\n",getpid());
      quit(CL_PIPEERR);
      return 0;
    }
    n += strlen(*message+n);
  }
  *(*message+*msg_len) = 0;
  return 1;
}

int main(int argc, char** argv, char** env)
{ int   n;
  char* p[32];
#ifndef HAVE_SIGSET
  struct sigaction sig_act;
#endif
  char  str[1024];

#ifndef HAVE_SIGSET
  sigfillset(&sig_act.sa_mask);
  sig_act.sa_flags = SA_RESTART;
  #define sigset(x,y) sig_act.sa_handler = y; sigaction(x,&sig_act,NULL)
#endif
  sigset(SIGHUP,signal_handler);
  sigset(SIGINT,signal_handler);
  sigset(SIGQUIT,signal_handler);
  sigset(SIGALRM,signal_handler);
  sigset(SIGPIPE,signal_handler);
  sigset(SIGTERM,signal_handler);
  setvbuf(stdout,NULL,_IOLBF,0);
  printf("SIN\n");
  scanf("%i %i %i %i %i %li %i %li %i %li %i %i %i %i ",&num_anon,&max_anon,
    &num_users,&max_users,&login_kick,&other_ip_s,&login_retries,&logc,&touch,&ipaddr,
    &sock,&fconv_mode,&chmodallowed,&checktimeout);
  get_string(&permission_file,str,sizeof(str));
  get_string(&readme_file,str,sizeof(str));
  get_string(&dupe_path,str,sizeof(str));
  get_string(&ff_path,str,sizeof(str));
  get_string(&script_path,str,sizeof(str));
  get_string(&pre_download,str,sizeof(str));
  get_string(&pre_upload,str,sizeof(str));
  if (!get_message("MSU",&message,&msg_len,&msg_size)) return close_reason;
  if (!get_message("MCU",&cdup_msg,NULL,NULL)) return close_reason;
  if (!get_message("MCD",&chdir_msg,NULL,NULL)) return close_reason;
  if (!get_message("MDE",&dele_msg,NULL,NULL)) return close_reason;
  if (!get_message("MDN",&download_msg,NULL,NULL)) return close_reason;
  if (!get_message("MLS",&list_msg,NULL,NULL)) return close_reason;
  if (!get_message("MMD",&mkdir_msg,NULL,NULL)) return close_reason;
  if (!get_message("MUP",&upload_msg,NULL,NULL)) return close_reason;
  if (!get_message("SSC",&shortcut,NULL,NULL)) return close_reason;
  get_group_list();
  get_user_list();
  get_cdpath();
  get_check();
  get_top();
  get_dupecheck();
  get_dshortcut(&dshortcut);
  if ((sockfr = fdopen(sock,"rt")) == NULL)
  { fprintf(stderr,"ftps(%i): socket fdopen failed\n");
    return CL_SOCKERR;
  }
  if ((sockf = fdopen(sock,"wt")) == NULL)
  { fprintf(stderr,"ftps(%i): socket fdopen failed\n");
    return CL_SOCKERR;
  }
  setvbuf(sockf,NULL,_IOLBF,0);
  if (send_reply(220,message) == -1) quit(CL_SOCKERR);
  while (!close_server)
  { client_activity("IDLE");
    handle_signals();
    if (!close_server) switch (read_ftp_command(str,sizeof(str),p,31,&n))
    { case -1 :
        if (errno && (errno != EINTR))
        { fprintf(stderr,"ftps(%i): read_ftp_command returned -1 (errno=%i)\n",
            getpid(),errno);
          quit(CL_SOCKERR);
        }
        break;
      case 1 :
        fprintf(stderr,"ftps(%i): read_ftp_command retured 1\n",getpid());
        break;
      default :
        handle_signals();
        if (!close_server)
        { if (dispatch_command(str,sizeof(str),p,n) == -1)
            if (errno != EINTR) quit(CL_SOCKERR);
        }
    }
  }
  switch (close_reason)
  { case CL_NORMAL :
      break;
    case CL_SHUTDOWN :
      send_reply(421,"Server shutdown, closing control connection.");
      break;
    case CL_TIMEOUT :
      send_reply(421,"Timeout, closing control connection.");
      break;
    case CL_ERROR :
      send_reply(421,"Server error, service not available.");
      break;
    case CL_KICK :
      send_reply(421,"Service no longer available for you.");
      break;
    case CL_SOCKERR :
      break;
    default :
      send_reply(421,"Service not available, closing control connection.");
  }
  printf("UUD %i\n",close_reason);
  fgets(str, sizeof(str), stdin);
  return close_reason;
}

/* EOF */
