/*   list.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996,1997,1998,1999 by Christoph Schwarz
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include "tweak.h"
#ifdef HAVE_SYS_SELECT_H
  #include <sys/select.h>
#endif
#ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
#endif
#include "externs.h"

#define LS_DOT       0x01
#define LS_LONG      0x02
#define LS_RECURSIVE 0x04
#define LS_TYPECHAR  0x08
#define LS_REVERSE   0x10
#define LS_DIRECTORY 0x20

#define SORT_MASK    0x0700
#define SORT_NONE    0x0000
#define SORT_NAME    0x0100
#define SORT_SIZE    0x0200
#define SORT_MTIME   0x0300
#define SORT_ATIME   0x0400
#define SORT_CTIME   0x0500

struct fileinfo_s
{ long size;
  long time;
  long nlink;
  long mode;
} ;

int wildcard_matching(char* wildcard, char* filename, int csens)
{ unsigned char  c1;
  unsigned char  c2;
  unsigned char  c3;
  unsigned char  c4;
  unsigned char* name = (unsigned char*)filename;
  unsigned char* wc   = (unsigned char*)wildcard;

  if ((readme_file && !strcmp(name,readme_file)) ||
    !strcmp(name,permission_file)) return 0;
  if (!*wc) return 1;
  while (*name && *wc)
  { switch (*wc)
    { case '?' : break;
      case '*' :
        if ((char*)name == filename)
        { if (*filename == '.') return 0;
          name++;
        }
        switch (*(++wc))
        { case '\0' : return 1;
          case '*' :
          case '?' : c1 = 0; c2 = 255; break;
          case '[' :
            if ((strlen(wc) < 5) || (*(wc+2) != '-') || (*(wc+4) != ']'))
              return 0;
            c1 = *(wc+1);
            c2 = *(wc+3);
            if (!csens)
            { c1 = toupper(c1);
              c2 = toupper(c2);
              c3 = tolower(c1);
              c4 = tolower(c2);
            }
            wc += 4;
            break;
          default  :
            c1 = c2 = *wc;
            if (!csens)
            { c1 = c2 = toupper(c1);
              c3 = c4 = tolower(c1);
            }
        }
        if (csens)
        { while (*name && ((*name < c1) || (*name > c2))) name++;
        } else
        { while (*name && ((*name < c1) || (*name > c2))
            && ((*name < c3) || (*name > c4))) name++;
        }
        if (!*name) return 0;
        break;
      case '[' :
         if ((strlen(wc) < 5) || (*(wc+2) != '-') || (*(wc+4) != ']') ||
           (*name < *(wc+1)) || (*name > *(wc+3))) return 0;
         wc += 4;
         break;
      default  :
        if (csens)
        { if (*wc != *name) return 0;
        } else
        { if (tolower(*wc) != tolower(*name)) return 0;
        }
    }
    name++;
    wc++;
  }
  if (csens)  return (*name == *wc) || (!strcmp(wc,"*"));
    else return (tolower(*name) == tolower(*wc)) || (!strcmp(wc,"*"));
}

void endfind(DIR* d)
{ if (d) closedir(d);
}

struct dirent* findnext(DIR* d, char* wc)
{ struct dirent* f;

  while ((f = readdir(d)) != NULL)
    if (wildcard_matching(wc,f->d_name,1)) return f;
  return f;
}

struct dirent* findfirst(DIR** d, char* dir, char* wc)
{ if ((*d = opendir(dir && *dir ? dir : ".")) == NULL) return NULL;
  return findnext(*d,wc);
}

int eval_options(char *p, int o)
{ while (*p)
  { switch (*p)
    { case 'A' :
      case 'a' : o |= LS_DOT; break;
      case 'c' : o = (o & ~SORT_MASK) | SORT_CTIME; break;
      case 'd' : o |= LS_DIRECTORY; break;
      case 'F' :
      case 'p' : o |= LS_TYPECHAR; break;
      case 'f' : o = (o & ~SORT_MASK) | SORT_NONE ; break;
      case 'l' : o |= LS_LONG; break;
      case 'R' : o |= LS_RECURSIVE; break;
      case 'r' : o |= LS_REVERSE; break;
      case 'S' : o = (o & ~SORT_MASK) | SORT_SIZE; break;
      case 't' : o = (o & ~SORT_MASK) | SORT_MTIME; break;
      case 'u' : o = (o & ~SORT_MASK) | SORT_ATIME; break;
    }
    p++;
  }
  return o;
}

int name_less(char* buffer, char* name, struct stat* st, int options)
{ struct fileinfo_s* fi;
  int                l;

  l = (strlen(buffer) + 8) & ~0x07;
  fi = (struct fileinfo_s*)(buffer+l);
  if (options & LS_REVERSE) switch (options & SORT_MASK)
  { case SORT_NAME  : return (strcmp(buffer,name) >= 0);
    case SORT_MTIME : return fi->time <= st->st_mtime;
    case SORT_ATIME : return fi->time <= st->st_atime;
    case SORT_CTIME : return fi->time <= st->st_ctime;
    case SORT_SIZE  : return fi->size <= st->st_size;
    default         : return 0;
  } else switch (options & SORT_MASK)
  { case SORT_NAME  : return (strcmp(buffer,name) <= 0);
    case SORT_MTIME : return fi->time >= st->st_mtime;
    case SORT_ATIME : return fi->time >= st->st_atime;
    case SORT_CTIME : return fi->time >= st->st_ctime;
    case SORT_SIZE  : return fi->size >= st->st_size;
    default         : return 1;
  }
}

int add_name(char** buffer, int* size, int* used, char* name, struct stat* st,
  int options)
{ struct fileinfo_s* fi;
  int                l   = strlen(name);
  int                ofs = 0;
  int                ofsd;

  if (*used+l+sizeof(struct fileinfo_s)+8 >= *size)
  { *size += 1024;
    *buffer = alloc(*buffer,*size);
    if (!*buffer) return -1;
  }
  while ((ofs < *used) && name_less(*buffer+ofs,name,st,options))
  { ofs += strlen(*buffer+ofs) + sizeof(struct fileinfo_s) + 8;
    ofs &= ~0x07;
  }
  ofsd = (ofs + l + sizeof(struct fileinfo_s) + 8) & ~0x07;
  if (ofs < *used) memmove(*buffer+ofsd,*buffer+ofs,*used-ofs);
  strcpy(*buffer+ofs,name);
  ofs += (l + 8) & ~0x07;
  fi = (struct fileinfo_s*)(*buffer+ofs);
  fi->size = st->st_size;
  switch (options & SORT_MASK)
  { case SORT_ATIME : fi->time = st->st_atime; break;
    case SORT_CTIME : fi->time = st->st_ctime; break;
    default         : fi->time = st->st_mtime; break;
  }
  fi->nlink = st->st_nlink;
  fi->mode = st->st_mode;
  *used += (l + sizeof(struct fileinfo_s) + 8) & ~0x07;
  return 0;
}

void permission_str(int dir, long perm, char* str)
{ int i;
  int p;

  if (dir) *(str++) = 'd'; else *(str++) = '-';
  for (i=0; i<3; i++)
  { p = (perm >> ((2-i)*8)) & 0xFF;
    if (dir)
    { if (p & P_LIST) *(str++) = 'r'; else *(str++) = '-';
      if (p & (P_MKDIR+P_WRITE)) *(str++) = 'w'; else *(str++) = '-';
      if (p & P_CHDIR) *(str++) =  'x'; else *(str++) = '-';
    } else
    { if (p & FP_READ) *(str++) = 'r'; else *(str++) = '-';
      if (p & FP_WRITE) *(str++) = 'w'; else *(str++) = '-';
      *(str++) = '-';
    }
  }
  *str = '\0';
}

void uid_gid_str(long uid, long gid, char* str)
{ char* name;

  name = get_username(uid);
  if (name) sprintf(str," %-9s",name); else sprintf(str," %-9li",uid);
  name = get_groupname(gid);
  if (name) sprintf(str+10," %-9s",name); else sprintf(str+10," %-9li",gid);
}

void file_info(struct fileinfo_s* st, long uid, long gid, long perm, char* str,
  char* dir, char* name, int year, struct file_s* fp, int n)
{ long           dgid;
  long           dlc;
  long           dperm;
  long           duid;
  struct file_s* f;
  int            found;
  struct tm*     tm;

  if (st->mode & S_IFDIR)
  { if (read_permission_file(dir,&duid,&dgid,&dperm,&dlc,NULL,NULL) == -1)
    { duid = dgid = -1;
      dlc = dperm = 0;
    }
    permission_str(1,dperm,str);
  } else
  { f = find_name(fp,name,n,&found);
    if (found)
    { duid = f->uid;
      dgid = f->gid;
      dlc = f->dlc;
      permission_str(0,f->perm,str);
    } else
    { duid = uid;
      dgid = gid;
      dlc = 0;
      permission_str(0,dirperm_to_fileperm(perm),str);
    }
  }
  sprintf(str+10," %3li",dlc);
  uid_gid_str(duid,dgid,str+strlen(str));
  sprintf(str+strlen(str)," %8li ",st->size);
  tm = localtime(&st->time);
  if (tm->tm_year == year) strftime(str+strlen(str),14,"%b %d %H:%M ",tm);
    else strftime(str+strlen(str),14,"%b %d  %Y ",tm);
}

long send_list_recursive(int ds, char* dir, char* name, char* wc, int options)
{ long           blocks;
  char*          buffer;
  DIR*           d;
  struct dirent* de;
  struct file_s* f;
  fd_set         fdset_r;
  fd_set         fdset_w;
  long           gid;
  long           l;
  int            n;
  int            nfiles;
  int            ofs;
  char*          p[8];
  long           perm;
  int            size;
  struct stat    st;
  char           str[128];
  struct timeval t;
  time_t         tim;
  long           uid;
  int            used;
  int            year;

  if (!superuser() && (permission_part_ok(dir,P_LIST,0) == -1))
    return send(ds,". unreadable\r\n",14,0);
  if (read_permission_file(dir,&uid,&gid,&perm,NULL,(char**)&f,&nfiles) == -1)
  { uid = gid = -1;
    perm = 0;
  }
  if (strcmp(dir,"/"))
  { d = opendir(dir+1);
    *(name++) = '/';
  } else d = opendir(".");
  if (!d) return -1;
  blocks = 0;
  buffer = NULL;
  size = used = 0;
  while ((de = readdir(d)) != NULL)
  { if (((options & LS_DOT) || (*de->d_name != '.')) &&
      wildcard_matching(wc,de->d_name,1))
    { strcpy(name,de->d_name);
      if (!stat(dir+1,&st))
      { blocks += st.st_blocks >> 1;
        if (add_name(&buffer,&size,&used,de->d_name,&st,options) == -1)
        { closedir(d);
          errno = ENOMEM;
          return -1;
        }
      }
    }
  }
  closedir(d);
  if ((options & LS_LONG) && (!wc || !*wc))
  { sprintf(str,"total %li\r\n",blocks);
    if (send(ds,str,strlen(str),0) == -1)
    { if (buffer) free(buffer);
      return -1;
    }
  }
  if (!buffer) return 0;
  time(&tim);
  year = localtime(&tim)->tm_year;
  FD_ZERO(&fdset_r);
  FD_ZERO(&fdset_w);
  ofs = 0;
  while (ofs < used)
  { FD_SET(sock,&fdset_r);
    FD_SET(ds,&fdset_w);
    t.tv_sec  = timeout;
    t.tv_usec = 0;
    if (select(FD_SETSIZE,&fdset_r,&fdset_w,NULL,&t) <= 0)
    { free(buffer);
      return -1;
    }
    if (FD_ISSET(sock,&fdset_r))
    { switch (read_ftp_command(sock,str,sizeof(str),&p,8,&n))
      { case -1 :
          free(buffer);
          return -1;
        case  0 :
          if (strcasecmp(*p,"ABOR"))
          { if (send_reply(sock,503,"Unexpected command.") == -1)
            { free(buffer);
              return -1;
            }
          } else
          { free(buffer);
            send_reply(sock,426,"Transfer interrupted on request.");
            return 1;
          }
      }
    } else
    { l = strlen(buffer+ofs) + 1;
      if (options & LS_LONG)
      { strcpy(name,buffer+ofs);
        file_info((struct fileinfo_s*)(buffer+((ofs+l+7) & ~0x07)),uid,gid,
          perm,str,dir,name,year,f,nfiles);
        if (send(ds,str,strlen(str),0) == -1)
        { free(buffer);
          return -1;
        }
      }
      if (send(ds,buffer+ofs,l-1,0) == -1) return -1;
      l = (l + 7) & ~0x07;
      if (options & LS_TYPECHAR)
      { blocks = ((struct fileinfo_s*)(buffer+ofs+l))->mode;
        if (blocks & S_IFDIR) n = send(ds,"/",1,0); else
        if (blocks & S_IFIFO) n = send(ds,"|",1,0); else n = 0;
        if (n == -1)
        { free(buffer);
          return -1;
        }
      }
      if (send(ds,"\r\n",2,0) == -1)
      { free(buffer);
        return -1;
      }
      ofs += l + sizeof(struct fileinfo_s);
    }
  }
  ofs = 0;
  if (options & LS_RECURSIVE) while (ofs < used)
  { l = (strlen(buffer+ofs) + 8) & ~0x07;
    if (strcmp(buffer+ofs,".") && strcmp(buffer+ofs,"..") &&
      (((struct fileinfo_s*)(buffer+ofs+l))->mode & S_IFDIR))
    { strcpy(name,buffer+ofs);
      n = send(ds,"\r\n",2,0);
      if (n != -1) n = send(ds,dir,strlen(dir),0);
      if (n != -1) n = send(ds,":\r\n",3,0);
      if (n != -1) n = send_list_recursive(ds,dir,dir+strlen(dir),wc,options);
      if (n == -1)
      { free(buffer);
        return -1;
      }
    }
    ofs += l + sizeof(struct fileinfo_s);
  }
  free(buffer);
  return blocks;
}

int do_list(char* fspec, int options)
{ int   ds;
  char* dir;
  int   n;
  char* name;
  char  wd[256];

  split_dir(fspec,&dir,&name,!(options & LS_DIRECTORY));
  n = add_dir(cwd,dir,wd,sizeof(wd));
  client_activity("LIST %s",n == -1 ? fspec : wd);
  client_log_start(LG_LS,"LIST %s",n == -1 ? fspec : wd);
  if ((n == -1) || (permission_ok(dir,P_LIST) == -1))
    return send_error(550,"list",LG_LS);
  if (pasvsock == -1)
  { if ((ds = socket(PF_INET,SOCK_STREAM,0)) == -1)
      return send_error(426,"socket",LG_LS);
    if (connect(ds,(struct sockaddr*)&addr,sizeof(addr)) == -1)
    { close(ds);
      return send_error(426,"connect",LG_LS);
    }
  } else
  { n = sizeof(addr);
    alarm(timeout);
    ds = accept(pasvsock,(struct sockaddr*)&addr,&n);
    alarm(0);
    if (ds == -1) return send_error(426,"accept",LG_LS);
  }
  if (send_reply(150,"Opening ASCII mode data connection for /bin/ls.") == -1)
  { close(ds);
    client_log_error(LG_LS);
    return -1;
  }
  if (send_list_recursive(ds,wd,wd+strlen(wd),name,options) == -1)
  { close(ds);
    return send_error(426,"list",LG_LS);
  }
  close(ds);
  client_log_finish(LG_LS,NULL);
  return send_message(226,1,list_msg,dir);
}

int cmd_list(char** p, int n, char* buffer, int bufferlen)
{ int i       = -1;
  int options = LS_LONG | SORT_NAME;

  if (n > 2) return send_reply(501,"Syntax: LIST [<options>] [<filespec>]");
  if (n >= 1) if (**p == '-')
  { if (n == 2) i = 1;
    options = eval_options(*p,options);
  } else i = 0;
  return do_list(i != -1 ? *(p+i) : "",options);
}

int cmd_nlst(char** p, int n, char* buffer, int bufferlen)
{ int i       = -1;
  int options = SORT_NAME;

  if (n > 2) return send_reply(501,"Syntax: NLST [<options>] [<filespec>]");
  if (n >= 1) if (**p == '-')
  { if (n == 2) i = 1;
    options = eval_options(*p,options);
  } else i = 0;
  return do_list(i != -1 ? *(p+i) : "",options);
}

/* EOF */
