/*   perm.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996,1997,1998,1999 by Christoph Schwarz
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <utime.h>
#include <sys/stat.h>
#include "tweak.h"
#include "externs.h"

int superuser(void)
{ return !(usr.class & 0x03);
}

int groupop(void)
{ return (usr.class & 0x100);
}

int useredit(void)
{ return (usr.class & 0x200);
}

int invisible(void)
{ return (usr.class & 0x400);
}

int coloruser(void)
{ return (usr.class & 0x80);
}

int in_group(long gid)
{ int i;

  if (usr.gid == gid) return 1;
  for (i=0; i<8; i++) if ((usr.gids[i] != -1) && (usr.gids[i] == gid))
    return 1;
  return 0;
}

long get_gid(char* name)
{ struct group_s* g = group;
  long            i;

  for (i=0; i<group_count; i++)
  { if (!strcasecmp(g->name,name)) return g->gid;
    g++;
  }
  if (sscanf(name,"%li",&i)) return i; else return -1;
}

long get_uid(char* name)
{ long           i;
  struct user_s* u = user;

  for (i=0; i<user_count; i++)
  { if (!strcasecmp(u->name,name)) return u->uid;
    u++;
  }
  if (sscanf(name,"%li",&i)) return i; else return -1;
}

struct group_s* get_group_by_gid(long gid)
{ struct group_s* g = group;
  int             i;

  for (i=0; i<group_count; i++, g++) if (g->gid == gid) return g;
  return NULL;
}

char* get_groupname(long gid)
{ struct group_s* g;
  static char*    incoming = "incoming";
  static char*    nogroup = "nogroup";

  if (gid == -1) return nogroup;
  if (gid == -2) return incoming;
  g = get_group_by_gid(gid);
  if (g) return g->name; else return NULL;
}

struct user_s* get_user_by_uid(long uid)
{ int            i = 0;
  int            j;
  int            k = user_count - 1;

  while (i <= k)
  { j = (i + k) >> 1;
    if ((user+j)->uid < uid) i = j + 1;
      else if ((user+j)->uid > uid) k = j - 1; else
    { while (j && ((user+j-1)->uid == uid)) j--;
      return user+j;
    }
  }
  return NULL;
}

char* get_username(long uid)
{ static char*   incoming = "incoming";
  static char*   nobody = "nobody";
  struct user_s* u;

  if (uid == -1) return nobody;
  if (uid == -2) return incoming;
  u = get_user_by_uid(uid);
  if (u) return u->name; else return NULL;
}

void split_dir(char* fp, char** dir, char** name, int testdir)
{ char*       p;
  struct stat s;
  char        wd[256];

  if (testdir) if (!add_dir(cwd,fp,wd,sizeof(wd)) &&
    !stat(strcmp(wd,"/") ? wd+1 : ".",&s))
  { if (s.st_mode & S_IFDIR)
    { *dir = fp;
      *name = fp+strlen(fp);
      return;
    }
  }
  if (*fp == '/')
  { p = strrchr(fp+1,'/');
    if (!p)
    { *dir = NULL;
      *name = fp+1;
    } else
    { *p = '\0';
      *dir = fp;
      *name = p+1;
    }
  } else
  { p = strrchr(fp,'/');
    if (!p)
    { *dir = fp+strlen(fp);
      *name = fp;
    } else
    { *p = '\0';
      *dir = fp;
      *name =  p+1;
    }
  }
}

int add_dir(char* src, char* dir, char* dest, int maxlen)
{ char  c;
  char* p = dir;
  char* q;

  if (!dir)
  { strcpy(dest,"/");
    return 0;
  }
  if (*dir == '/')
  { strcpy(dest,"/");
    dir++;
    p++;
  } else
  { if (strlen(src) >= maxlen)
    { errno = ENAMETOOLONG;
      return -1;
    }
    if (dest != src) strcpy(dest,src);
  }
  while (*dir)
  { while (*p && *p!='/') p++;
    c = *p;
    *p = '\0';
    if (!strcmp(dir,".."))
    { if (!*dest || !strcmp(dest,"/"))
      { errno = ENOENT;
        return -1;
      }
      q = dest+strlen(dest)-1;
      while (q>dest && *q!='/') q--;
      if ((q == dest) && (*q == '/')) *(q+1) = '\0'; else *q = '\0';
    } else if (strcmp(dir,".") && strcmp(dir,"/"))
    { if (strlen(dest)+strlen(dir)+1 >= maxlen)
      { errno = ENAMETOOLONG;
        return -1;
      }
      if (strcmp(dest,"/")) strcat(dest,"/");
      strcat(dest,dir);
    }
    if (c) *(p++) = c;
    dir = p;
  }
  return 0;
}

FILE* file_open(char* dir, char* filename, const char* mode, int pcheck)
{ struct stat    st;
  struct utimbuf t;
  char           wd[256];

  if (add_dir(cwd,dir,wd,sizeof(wd)) == -1) return NULL;
  if (strlen(wd)+strlen(filename)+1 >= sizeof(wd))
  { errno = ENAMETOOLONG;
    return NULL;
  }
  if (strcmp(wd,"/")) strcat(wd,"/");
  strcat(wd,filename);
  if (pcheck && !superuser())
  { if (!strcmp(filename,permission_file))
    { errno = EACCES;
      return NULL;
    }
    if (readme_file && !strcmp(filename,readme_file))
    { errno = EACCES;
      return NULL;
    }
    if (pcheck == P_READ)
    { if (stat(wd+1,&st) == -1) return NULL;
      if (permission_ok(dir,P_CHDIR) == -1) return NULL;
      if (file_permission_ok(dir,filename,FP_READ) == -1) return NULL;
    } else
    { if (stat(wd+1,&st) == -1)
      { if (permission_ok(dir,P_WRITE) == -1) return NULL;
      } else
      { if (permission_ok(dir,P_CHDIR) == -1) return NULL;
        if (file_permission_ok(dir,filename,FP_WRITE) == -1) return NULL;
      }
    }
  }
  if (touch)
  { t.modtime = time(&t.actime);
    utime(wd+1,&t);
  }
  return fopen(wd+1,mode);
}

int read_cdpath(char* dir, long* uid, long* gid, long* perm)
{ int  l;
  int  ofs = 0;
  char wd[256];

  if (add_dir(cwd,dir,wd,sizeof(wd)) == -1) return -1;
  while (ofs < cdpath_size)
  { l = strlen(cdpath+ofs);
    if (!strncmp(wd,cdpath+ofs,l))
    { ofs += (l + 8) & ~0x07;
      *uid = *(long*)(cdpath+ofs);
      *gid = *(long*)(cdpath+ofs+sizeof(long));
      *perm = *(long*)(cdpath+ofs+(sizeof(long)*2));
      return 0;
    }
    ofs += ((l + 8) & ~0x07) + (sizeof(long) * 3);
  }
  return -1;
}

int read_permission_file(char* dir, long* uid, long* gid, long* perm,
  long* dlc, char** buffer, int* n)
{ FILE*         file;
  char          line[32];
  struct perm_s p;
  struct stat   st;

  if (dlc) *dlc = 1;
  if (buffer) *buffer = NULL;
  if (n) *n = 0;
  if (read_cdpath(dir,uid,gid,perm) == 0) return 0;
  if ((file = file_open(dir,permission_file,"rb",0)) == NULL) return -1;
  if (fstat(fileno(file),&st) == -1)
  { fclose(file);
    return -1;
  }
  if (st.st_size < sizeof(struct perm_s))
  { if (buffer && n)
    { *n = 0;
      if ((*buffer = malloc(sizeof(struct file_s))) == NULL)
      { fclose(file);
        return -1;
      }
    }
    if (fread(line,1,st.st_size,file) != st.st_size)
    { free(*buffer);
      fclose(file);
      return -1;
    }
    if (sscanf(line,"%li %li %li",uid,gid,perm) != 3)
    { free(*buffer);
      fclose(file);
      return -1;
    }
    if (dlc) *dlc = 0;
  } else
  { if (!fread(&p,sizeof(p),1,file))
    { fclose(file);
      return -1;
    }
    *uid = p.uid;
    *gid = p.gid;
    *perm = p.perm;
    if (dlc) *dlc = p.dlc;
    if (buffer && n)
    { st.st_size -= sizeof(p);
      if ((*buffer = malloc(st.st_size+sizeof(struct file_s))) == NULL)
      { fclose(file);
        return -1;
      }
      if (fread(*buffer,1,st.st_size,file) != st.st_size)
      { free(*buffer);
        fclose(file);
        return -1;
      }
      *n = st.st_size / sizeof(struct file_s);
    }
  }
  fclose(file);
  return 0;
}

int remove_permission_file(char* dir)
{ char wd[256];

  if (strlen(dir)+strlen(permission_file) >= sizeof(wd))
  { errno = ENAMETOOLONG;
    return -1;
  }
  if (strcmp(dir,"/")) sprintf(wd,"%s/%s",dir+1,permission_file); else
    strcpy(wd,permission_file);
  unlink(wd);
  if (readme_file)
  { if (strcmp(dir,"/")) sprintf(wd,"%s/%s",dir+1,readme_file); else
      strcpy(wd,readme_file);
    unlink(wd);
  }
  return 0;
}

int write_cdpath(char* dir, long uid, long gid, long perm)
{ int  l;
  int  ofs = 0;
  int  saveofs;
  char wd[256];

  if (add_dir(cwd,dir,wd,sizeof(wd)) == -1) return -1;
  while (ofs < cdpath_size)
  { l = strlen(cdpath+ofs);
    if (!strncmp(wd,cdpath+ofs,l))
    { if (l == strlen(wd))
      { saveofs = ofs;
        ofs += (l + 8) & ~0x07;
        *(long*)(cdpath+ofs) = uid;
        *(long*)(cdpath+ofs+sizeof(long)) = gid;
        *(long*)(cdpath+ofs+sizeof(long)*2) = perm;
        printf("SCC %s %li %li %li\n",cdpath+saveofs,uid,gid,perm);
        return 0;
      } else
      { errno = EACCES;
        return -1;
      }
    }
    ofs += ((l + 8) & ~0x07) + (sizeof(long)*3);
  }
  errno = 0;
  return -1;
}

int write_permission_file(char* dir, long uid, long gid, long perm)
{ int           e;
  FILE*         file;
  struct perm_s p;
  struct stat   st;

  if (write_cdpath(dir,uid,gid,perm) == -1)
  { if (errno) return -1;
  } else return 0;
  if ((file = file_open(dir,permission_file,"r+b",0)) == NULL)
    if ((file = file_open(dir,permission_file,"w+b",0)) == NULL) return -1;
  if (fstat(fileno(file),&st) == -1) st.st_size = 0;
  p.uid = uid;
  p.gid = gid;
  p.perm = perm;
  if (st.st_size < sizeof(p))
  { p.dlc = p.dlct = 0;
    memset(&p.reserved,0,sizeof(p.reserved));
  }
  if (!fwrite(&p,st.st_size < sizeof(p) ? sizeof(p) : sizeof(long)*3,1,file))
  { e = errno;
    fclose(file);
    errno = e;
    return -1;
  }
  fclose(file);
  return 0;
}

int permission_part_ok(char* dir, int p, int uidok)
{ long gid;
  long perm;
  long uid;

  if (read_permission_file(dir,&uid,&gid,&perm,NULL,NULL,NULL) == -1)
  { errno = EACCES;
    return -1;
  }
/*  if (uidok) if (usr.uid == uid) return 0;*/
  errno = EACCES;
  if (p & P_CHMOD) if (usr.uid != uid) return -1;
  if (usr.uid == uid) if (((perm >> 16) & p) == p) return 0;
  if (in_group(gid)) if (((perm >> 8) & p) == p) return 0;
  if ((perm & p) == p) return 0;
  return -1;
}

int permission_ok(char* dir, int perm)
{ char        c;
  char*       p;
  struct stat st;
  char        wd[256];

  if ((perm<0) || (perm>255))
  { errno = EACCES;
    return -1;
  }
  if (!dir) strcpy(wd,"/"); else
    if (add_dir(cwd,dir,wd,sizeof(wd)) == -1) return -1;
  if (strcmp(wd,"/"))
  { if (stat(wd+1,&st) == -1) return -1;
    if (!S_ISDIR(st.st_mode))
    { errno = ENOTDIR;
      return -1;
    }
  }
  if (superuser()) return 0;
  p = wd+1;
  do
  { while (*p && (*p != '\0')) p++;
    c = *p;
    *p = '\0';
    if (c && *(p+1))
    { if (permission_part_ok(wd,P_CHDIR,0) == -1) return -1;
      *(p++) = c;
    } else
    { if (permission_part_ok(wd,perm,0) == -1) return -1;
      if (c) p++;
    }
  } while (*p);
  return 0;
}

struct file_s* find_name(struct file_s* buffer, char* name, int n, int* found)
{ int i = 0;
  int j;
  int k = n-1;

  while (i <= k)
  { j = (i + k) >> 1;
    n = strcmp(name,(buffer+j)->name);
    if (n < 0) k = j - 1; else if (n > 0) i = j + 1; else
    { *found = 1;
      return buffer+j;
    }
  }
  *found = 0;
  return buffer+i;
}

long dirperm_to_fileperm(long perm) 
{ int  i;
  long p = 0x01000000;

  for (i=0; i<=2; i++)
  { if ((perm >> ((2-i)*8)) & P_READ) p |= FP_READ << (2-i)*8;
    if ((perm >> ((2-i)*8)) & P_OVERWRITE) p |= FP_WRITE << (2-i)*8;
    if (((perm >> ((2-i)*8)) & P_CHMOD) && !i) p |= FP_CHMOD << (2-i)*8;
  }
  return p;   
}

int read_file_permission(char* dir, char* name, long* uid, long* gid,
  long* perm)
{ char*          buffer;
  struct file_s* f;
  int            found;
  int            n;

  if (read_permission_file(dir,uid,gid,perm,NULL,&buffer,&n) == -1)
  { errno = EACCES;
    return -1;
  }
  f = find_name((struct file_s*)buffer,name,n,&found);
  if (found)
  { *uid  = f->uid;
    *gid  = f->gid;
    if (*perm) *perm = f->perm; else *perm = dirperm_to_fileperm(*perm);
  } else *perm = dirperm_to_fileperm(*perm);
  free(buffer);
  return 0;
}

int file_permission_ok(char* dir, char* name, long tperm)
{ long gid;
  long perm;
  long uid;

  if ((0 == strcmp(permission_file,name)) || (0 == strcmp(readme_file,name))) return -1;
  if (superuser()) return 0;
  if (read_file_permission(dir,name,&uid,&gid,&perm) == -1) return -1;
  if (uid == usr.uid) if ((((perm >> 16) & FP_ALL) & tperm) == tperm)
    return 0;
  if (in_group(gid)) if ((((perm >> 8) & FP_ALL) & tperm) == tperm)
    return 0;
  if (((perm & FP_ALL) & tperm) == tperm) return 0;
  errno = EACCES;
  return -1;
}

/* EOF */
