/*
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996,1997,1998,1999 by Christoph Schwarz
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <syslog.h>
#include "tweak.h"
#ifdef DES_ENCRYPTION
  #include "des.h"
#endif
#include "externsd.h"

struct user_s* insert_user(struct user_s*);

char salt_a[] =
  "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ./";

int anonuser(struct user_s* u)
{ if (!u) return 0;
  return ((u->class & UC_MASK) == UC_ANON);
}

int normaluser(struct user_s* u)
{ if (!u) return 0;
  return ((u->class & UC_MASK) == UC_NORMAL);
}

int superuser(struct user_s* u)
{ if (!u) return 0;
  return ((u->class & UC_MASK) == UC_ROOT);
}

int allowed(struct user_s* u, long right)
{ if (superuser(u)) return 1;
  if (groupop(u) && (groupop_rights & right)) return 1;
  if (useredit(u) && (useredit_rights & right)) return 1;
  return 0;
}

int templateuser(struct user_s* u)
{ if (!u) return 0;
  return ((u->class & UC_MASK) == UC_TEMPLATE);
}

int coloruser(struct user_s* u)
{ if (!u) return 0;
  return (u->class & UE_COLOR);
}

int invisible(struct user_s* u)
{ if (!u) return 0;
  return (u->class & UE_INVISIBLE);
}

int no_pass_change(struct user_s* u)
{ if (!u) return 0;
  return (u->class & UE_NOPASSCH);
}

int account_disabled(struct user_s* u)
{ return u->class & 0x08;
}

int groupop(struct user_s* u)
{ return ((u->class & USC_MASK) == USC_GROUPOP);
}

int groupopof(struct user_s* g, struct user_s* u)
{ int i;
  if (!groupop(g)) return 0;
  if (g->gid==-1) return 0;
  if (u->gid==g->gid) return 1;
  for (i=0;i<8;i++)
    if (u->gids[i]==g->gid) return 1;
  return 0;
}

int useredit(struct user_s* u)
{ return ((u->class & USC_MASK) == USC_USEREDIT);
}

long get_gid(char* name)
{ struct group_s* g = group;
  long            i;

  for (i=0; i<group_count; i++)
  { if (!strcasecmp(g->name,name)) return g->gid;
    g++;
  }
  if (sscanf(name,"%li",&i)) return i; else return -1;
}

struct group_s* get_group_by_gid(long gid)
{ struct group_s* g = group;
  int             i;

  for (i=0; i<group_count; i++, g++) if (g->gid == gid) return g;
  return NULL;
}

struct group_s* get_group_by_name(char* name)
{ struct group_s* g = group;
  long            gid;
  int             i;

  for (i=0; i<group_count; i++, g++) if (!strcasecmp(g->name,name)) return g;
  if (sscanf(name,"%li",&gid)) return get_group_by_gid(gid);
  return NULL;
}

int get_groupname(long gid, char* name)
{ struct group_s* g = group;
  int             i;

  if (gid == -1)
  { strcpy(name,"nogroup");
    return 1;
  }
  for (i=0; i<group_count; i++, g++) if (g->gid == gid)
  { strcpy(name,g->name);
    return 1;
  }
  sprintf(name,"%li",gid);
  return 0;
}

struct user_s* get_user_by_uid(long uid)
{ int            i = 0;
  int            j;
  int            k = user_count - 1;

  while (i <= k)
  { j = (i + k) >> 1;
    if ((user+j)->uid < uid) i = j + 1; else if ((user+j)->uid > uid) k = j - 1;
      else return user+j;
  }
  return NULL;
}

struct user_s* get_user_by_name(char* name)
{ int            i;
  long           uid;
  struct user_s* u = user;

  for (i=0; i<user_count; i++, u++) if (!strcasecmp(u->name,name)) return u;
  if (sscanf(name,"%li",&uid)) return get_user_by_uid(uid);
  return NULL;
}

int parse_command(char* p, char** param, int max, int* n)
{ int q;

  while (skipspace(&p))
  { if (*n >= max) return 1;
    *(param+(*n)) = p;
    (*n)++;
    q = 0;
    while (!(q & 0x04)) switch (*p)
    { case '"'  :
        if (q & 0x02) p++; else
        { q ^= 0x01;
          memmove(p,p+1,strlen(p));
        }
        break;
      case '\'' :
        if (q & 0x01) p++; else
        { q ^= 0x02;
          memmove(p,p+1,strlen(p));
        }
        break;
      case '\\' :
        memmove(p,p+1,strlen(p));
        if (*p) p++;
        break;
      case '\t' :
      case ' ' :
        if (q) p++; else q |= 0x04;
        break;
      case '\0' :
        q |= 0x04;
        break;
      default :
        p++;
        break;
    }
    if (*p) *(p++) = '\0';
  }
  return 0;
}

int read_command(struct child_s* c, char* buffer, int bufferlen, char** param,
  int max, int* n)
{ char* p;
  int   q;

  *n = 0;
  if (!fgets(buffer,bufferlen,c->rdp)) return -1;
  q = strlen(buffer);
  if (!q) return 2;
  if (*(buffer+q-1) == '\n') q--;
  if (q && (*(buffer+q-1) == '\r')) q--;
  *(buffer+q) = '\0';
  p = buffer;
  if (!skipspace(&p)) return 4;
  *param = p;
  *n = 1;
  if (skipnspace(&p)) *(p++) = '\0'; else return 0;
  if (p == buffer+5)
  { *n = 2;
    *(param+1) = p;
    *(buffer+3) = '\0';
    return 0;
  }
  return parse_command(p,param,max-1,n);
}

void server_reply(struct child_s* c, const char* reply)
{ fprintf(c->wrp,"%s",reply);
  if (*(reply+strlen(reply)-1) != '\n') fprintf(c->wrp,"\n");
#ifdef DEBUG
  fprintf(debugf,"Server reply: %s",reply);
  if (*(reply+strlen(reply)-1) != '\n') fprintf(debugf,"\n");
#endif
  fflush(c->wrp);
}


void server_reply_lines(struct child_s* c, char* buffer)
{ char* p = buffer;
  char  line[1024];

  do
  { while (*p && (*p != '\n')) p++;
    if (*p) p++;
    line[0]='+';
    memcpy(line+1,buffer,p-buffer);
    line[p-buffer+1]='\0';
    server_reply(c,line);
    buffer = p;
  } while (*p);
}

void account_disable(struct child_s* c, char** p, int n)
{ struct user_s* u;

  if (!allowed(c->user,R_DISABLE))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n != 1)
  { server_reply(c,"-Syntax: ADI <username>");
    return;
  }
  u = get_user_by_name(*p);
  if (!u)
  { server_reply(c,"-User does not exist.");
    return;
  }
  if (groupop(c->user)  && !groupopof(c->user,u))
  { server_reply(c,"-User is not in your group.");
    return;
  }
  if (u->class & 0x08)
  { server_reply(c,"-Account is already disabled.");
    return;
  }
  u->class |= 0x08;
  server_reply(c,"+Account disabled.");
}

void account_enable(struct child_s* c, char** p, int n)
{ struct user_s* u;

  if (!allowed(c->user,R_ENABLE))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n != 1)
  { server_reply(c,"-Syntax: AEN <username>");
    return;
  }
  u = get_user_by_name(*p);
  if (!u)
  { server_reply(c,"-User does not exist.");
    return;
  }
  if (groupop(c->user)  && !groupopof(c->user,u))
  { server_reply(c,"-User is not in your group.");
    return;
  }
  if (!(u->class & 0x08))
  { server_reply(c,"-Account is already enabled.");
    return;
  }
  u->class ^= 0x08;
  server_reply(c,"+Account enabled.");
}

void add_user(struct child_s* c, char** p, int n)
{ struct user_s*  u;
  struct user_s   usr;
  struct group_s* g;
  int 	i,j;
  char  salt[2];
  int   temp;

  if (!allowed(c->user,R_USER_ADD))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n < 5)
  { server_reply(c,"-Syntax: ADDUSER <name> <group|gid> <password> <1:ratio|limit|leech> <ip0> [<ip1> [<ip2> [...]]]");
    return;
  }
  
  /* Name */
  if (strlen(*p) >= sizeof(usr.name))
  { server_reply(c,"-User name too long");
    return;
  }
  strcpy(usr.name,*p);

  /* Group */
  g = get_group_by_name(*(p+1));
  if (g==NULL) 
  { if (sscanf(*(p+1),"%i",&temp)) g = get_group_by_gid(temp);
    if (g==NULL)
    { server_reply(c,"-Group or GID not found.");
      return;
    }
  }
  usr.gid=g->gid;
  
  /* Password */
  *salt = salt_a[time(NULL) % 64];
  *(salt+1) = salt_a[(time(NULL) / 64) % 64];
  strcpy(usr.password,crypt(*(p+2),salt));

  /* Limit */
  if (!allowed(c->user,R_USER_CHANGE_RATIO))
  { if (groupop(c->user)) usr.limit = groupop_limit;
    else if (useredit(c->user)) usr.limit = useredit_limit;
    /* Should never happen */
    else usr.limit = 33;
  } else
  { if (!strcasecmp(*(p+3),"leech")) usr.limit = 0; 
    else if (sscanf(*(p+3),"1:%i",&temp)) usr.limit = 100 / temp;
    else if (sscanf(*(p+3),"%i",&temp)) usr.limit = temp;
    else
    { server_reply(c,"-Invalid limit. Should be 'leech', '1:xx' (ratio) or 'xx' (limit).");
      return;
    }
  }

  /* IPs */
  for (j = 0; j < 10; j++) memset(usr.idents[j],'\0',sizeof(usr.idents[j]));
  for (i = 0; i < (n-4); i++) if (get_iprange(p+4+i,usr.idents[i]) != 0) 
  { server_reply(c,"-Invalid IP");
    return;
  }

  /* Others */
  usr.uid           = 0;
  usr.kb_downloaded = 0;
  usr.kb_uploaded   = 0;
  usr.last_login    = 0;
  strcpy(usr.info,"");
  usr.credit = 0;
  usr.class = 0x10;
  memset(usr.gids,-1,sizeof(usr.gids));
  usr.class |= UC_NORMAL;
  u = user;
  for (n=0; n<user_count; n++, u++)
  { if (u->uid >= usr.uid) usr.uid = u->uid + 1;
    if (!strcasecmp(usr.name,u->name))
    { server_reply(c,"-User already exists.");
      return;
    }
  }
  if (!insert_user(&usr))
  { server_reply(c,"-Not enough memory.");
    return;
  }
  top_ok = 0;
  server_reply(c,"+User added.");
}


void color(struct child_s* c, char** p, int n)
{
  /* check if color is disabled */
  if (!(c->user->class & UE_COLOR))
  { 
    c->user->class |= UE_COLOR;
    server_reply(c,"+Colormode is now !Denabled!0.");
    return;
  }
  /* check if color is enabled */
  if (c->user->class & UE_COLOR)
  { 
    c->user->class ^= UE_COLOR;
    server_reply(c,"+Colormode is now disabled.");
    return;
  }
}

void count(struct child_s* c, char** p, int n)
{ char str[128];
  struct child_s *u;
  int nb=0;
  int i;

  u = child;
  for (i=0; i<children; i++, u++) if (u && u->user && !invisible(u->user)) nb++;
  sprintf(str,"+%i user%s connected.",nb,nb>1? "s" : "");
  server_reply(c,str);
}

void server_version(struct child_s* c, char** p, int n)
{ char str[128];
  sprintf(str,"+The server is running with FTP4ALL "VERSION".\n");
  server_reply(c,str);
}

void bind_port(struct child_s* c, char** p, int n)
{ struct sockaddr_in addr;
  struct linger      li = { 0, 0 };
  int                new_ss2;

  if (!superuser(c->user))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if ((n != 1) || !sscanf(*p,"%i",&new_ss2))
  { server_reply(c,"-Syntax: BIN <port>");
    return;
  }
  if (ss2 != -1)
  { close(ss2);
    ss2 = -1;
  }
  if (!new_ss2)
  { server_reply(c,"+Second port successfully unbound.");
    return;
  }
  ss2 = socket(PF_INET,SOCK_STREAM,0);
  if (ss2 == -1)
  { perror("ftpd: socket");
    server_reply(c,"-Unable to get socket.");
    return;
  }
  if (fcntl(ss2,F_SETFD,1) < 0) perror("ftpd: fcntl");
  if (setsockopt(ss2,SOL_SOCKET,SO_LINGER,(void *)&li,sizeof(li)) < 0)
    perror("ftpd: setsockopt");
  addr.sin_family      = AF_INET;
  if (other_ip == NULL) { addr.sin_addr.s_addr = htonl(INADDR_ANY); }
     else { addr.sin_addr.s_addr = inet_addr(other_ip); }
  addr.sin_port        = htons((u_short)new_ss2);
  if (bind(ss2,(struct sockaddr*)&addr,sizeof(addr)) == -1)
  { perror("ftpd: bind");
    close(ss2);
    ss2 = -1;
    server_reply(c,"-Unable to bind to port.");
    return;
  }
  listen(ss2,5);
  server_reply(c,"+Second port successfully allocated.");
}

void print_iprange_part(char* s, char l, char h, char c)
{ if (l == h) sprintf(s,"%i",(int)(unsigned char)l); else
  if ((l == 0) && ((unsigned char)h == 0xFF)) sprintf(s,"*"); else
    sprintf(s,"%i-%i",(int)(unsigned char)l,(int)(unsigned char)h);
  sprintf(s+strlen(s),"%c",c);
}

void print_iprange(char* s, char* iprange)
{ print_iprange_part(s,*(iprange+12),*(iprange+13),'.');
  print_iprange_part(s+strlen(s),*(iprange+14),*(iprange+15),'.');
  print_iprange_part(s+strlen(s),*(iprange+16),*(iprange+17),'.');
  print_iprange_part(s+strlen(s),*(iprange+18),*(iprange+19),'\0');
}

void print_ident(char* s, char* ident)
{ if (*ident!='\0') sprintf(s,"%s@",ident); else sprintf(s,"");
  print_iprange(s+strlen(s),ident);
}

#ifdef DES_ENCRYPTION
int write_user_file_des(const char* filename)
{ char*        buffer;
  C_Block      c;
  FILE*        file;
  int          n;
  C_Block      out;
  size_t       result;
  Key_schedule s;

  if ((file = fopen(filename,"wb")) == NULL) return -1;
  c = user_key;
  des_set_key(&c,&s);
  buffer = (char*)user;
  n = user_count * sizeof(struct user_s);
  while (n >= 8)
  { des_cbc_encrypt((C_Block*)buffer,&out,8,&s,&c,DES_ENCRYPT);
    memmove(buffer,&out,8);
    buffer += 8;
    n -= 8;
  }
  result = fwrite(user,sizeof(struct user_s),user_count,file);
  fclose(file);
  c = user_key;
  des_set_key(&c,&s);
  buffer = (char*)user;
  n = user_count * sizeof(struct user_s);
  while (n >= 8)
  { des_cbc_encrypt((C_Block*)buffer,&out,8,&s,&c,DES_DECRYPT);
    memmove(buffer,&out,8);
    buffer += 8;
    n -= 8;
  }
  if (result != user_count) return -1;
  return 0;
}
#endif

int write_user_file(const char* filename, FILE* file)
{ int            first;
  int            i;
  int            j;
  char           iprange[50];
  struct user_s* u = user;

  if (filename)
  {
#ifdef DES_ENCRYPTION
    if (user_encrypted) return write_user_file_des(filename);
#endif
    if ((file = fopen(filename,"wt")) == NULL) return -1;
  }
  fprintf(file,"# %i user(s)\n",user_count);
  for (i=0; i<user_count; i++, u++)
  { fprintf(file,"%s:%li:%li:%s:%li:%li:%li",u->name,u->uid,u->gid,
      u->password,u->kb_downloaded,u->kb_uploaded,u->last_login);
    first = 1;
    for (j=0; j<10; j++) if (memcmp(u->idents[j]+12,"\0\0\0\0\0\0\0\0",8))
    { print_ident(iprange,u->idents[j]);
      fprintf(file,"%c%s",first ? ':' : ',' ,iprange);
      first = 0;
    }
    if (first) fprintf(file,":0.0.0.0");
    fprintf(file,":%li:%s:0x%02lX:%li",u->limit,u->info,u->class,u->credit);
    first = 1;
    for (j=0; j<8; j++) if (u->gids[j] != -1)
    { fprintf(file,"%c%li",first ? ':' : ',' ,u->gids[j]);
      first = 0;
    }
    fprintf(file,"\n");
  }
  if (filename) fclose(file);
  return 0;
}

void dump_user_file(struct child_s* c, char** p, int n)
{ if (!superuser(c->user))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n != 1)
  { server_reply(c,"-Syntax: DUF <filename>");
    return;
  }
  if (write_user_file(*p,NULL) == -1)
    server_reply(c,"-Error writing user file.");
  else server_reply(c,"+User file saved successful.");
}

int write_group_file(const char* filename)
{ FILE*           file;
  struct group_s* g;
  int             i;

  if ((file = fopen(filename,"wt")) == NULL) return -1;
  fprintf(file,"# %i group(s)\n",group_count);
  g = group;
  for (i=0; i<group_count; i++)
  { if (fprintf(file,"%s:%li\n",g->name,g->gid) < 0)
    { fclose(file);
      return -1;
    }
    g++;
  }
  fclose(file);
  return 0;
}

void dump_group_file(struct child_s* c, char** p, int n)
{ if (!superuser(c->user))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n != 1)
  { server_reply(c,"-Syntax: DGF <filename>");
    return;
  }
  if (write_group_file(*p) == -1) server_reply(c,"-Error writing group file.");
    else server_reply(c,"+Group file saved successful.");
}

int read_permfile(char* filename, char** buffer, int* n)
{ FILE*          file;
  struct perm_s* pe;
  struct stat    st;

  if (stat(filename,&st) == -1) st.st_size = 0;
  if (st.st_size < sizeof(struct perm_s))
  { if ((*buffer = malloc(sizeof(struct perm_s)+sizeof(struct file_s))) == NULL)
      return -1;
    pe = (struct perm_s*)*buffer;
    if ((file = fopen(filename,"rt")) != NULL)
    { if (fscanf(file,"%li %li %li",&pe->uid,&pe->gid,&pe->perm) != 3)
      { free(*buffer);
        fclose(file);
        return -1;
      }
      fclose(file);
    } else
    { pe->uid = pe->gid = -1;
      pe->perm = umask_root;
    }
    pe->dlc = pe->dlct = 0;
    memset(&pe->reserved,0,sizeof(pe->reserved));
    *n = 0;
  } else
  { if ((*buffer = malloc(st.st_size+sizeof(struct file_s))) == NULL) return -1;
    if ((file = fopen(filename,"rb")) == NULL)
    { free(*buffer);
      return -1;
    }
    if (fread(*buffer,1,st.st_size,file) != st.st_size)
    { free(*buffer);
      fclose(file);
      return -1;
    }
    fclose(file);
    *n = (st.st_size - sizeof(struct perm_s)) / sizeof(struct file_s);
  }
  return 0;
}

int write_permfile(char* filename, char** buffer, int n)
{ FILE* file;

  if ((file = fopen(filename,"wb")) == NULL)
  { free(*buffer);
    return -1;
  }
  n = (n * sizeof(struct file_s)) + sizeof(struct perm_s);
  if (fwrite(*buffer,1,n,file) != n)
  { free(*buffer);
    fclose(file);
    return -1;
  }
  free(*buffer);
  fclose(file);
  return 0;
}

struct file_s* find_name(struct file_s* buffer, char* name, int n, int* found)
{ int i = 0;
  int j;
  int k = n-1;

  while (i <= k)
  { j = (i + k) >> 1;
    n = strcmp(name,(buffer+j)->name);
    if (n < 0) k = j - 1; else if (n > 0) i = j + 1; else
    { *found = 1;
      return buffer+j;
    }
  }
  *found = 0;
  return buffer+i;
}

long user_umask(struct user_s* u)
{ long perm;

  if (superuser(u)) perm = umask_root; else
    if (anonuser(u)) perm = umask_anon; else perm = umask_user;
  return perm;
}

long dirperm_to_fileperm(long perm)
{ int  i;
  long p = 0x01000000;

  for (i=0; i<=2; i++)
  { if ((perm >> ((2-i)*8)) & P_READ) p |= FP_READ << (2-i)*8;
    if ((perm >> ((2-i)*8)) & P_OVERWRITE) p |= FP_WRITE << (2-i)*8;
    if (((perm >> ((2-i)*8)) & P_CHMOD) && !i) p |= FP_CHMOD << (2-i)*8;
  }
  return p;
}

char* get_filename(char* path)
{ char* p;

  p = path+strlen(path);
  while ((p > path) && (*p != '/')) p--;
  return *p == '/' ? p+1 : p;
}

void file_chmod(struct child_s* c, char** p, int n)
{ char*          buffer;
  char           cmd;
  long           dperm;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  long           mode;
  struct perm_s* pe;
  int            who;

  sscanf(*p,"%i %c %li",&who,&cmd,&mode);
  skipnspace(p);
  skipspace(p);
  skipnspace(p);
  skipspace(p);
  skipnspace(p);
  (*p)++;
  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
      strcpy(f->name,filename);
      f->uid = pe->uid;
      f->gid = pe->gid;
      f->dlc = 0;
      f->perm = dirperm_to_fileperm(pe->perm);
      n++;
    }
    switch (cmd)
    { case '+' :
      case '-' :
        dperm = 0;
        if (who & 0x01) dperm |= mode<<16;
        if (who & 0x02) dperm |= mode<<8;
        if (who & 0x04) dperm |= mode;
        if (cmd == '+') f->perm |= dperm; else f->perm &= ~dperm;
        break;
      case '=' :
        if (who & 0x01) f->perm = (f->perm & 0x00FFFF) | (mode<<16);
        if (who & 0x02) f->perm = (f->perm & 0xFF00FF) | (mode<<8);
        if (who & 0x04) f->perm = (f->perm & 0xFFFF00) | mode;
        break;
    }
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+");
      return;
    }
  }
  server_reply(c,"-");
}

void file_chown(struct child_s* c, char** p, int n)
{ char*           buffer;
  struct file_s*  f;
  struct file_s*  f_start;
  char            filename[sizeof(f->name)];
  char*           fname_start;
  int             found;
  long            gid;
  struct perm_s*  pe;
  long            uid;

  sscanf(*p,"%li %li",&uid,&gid);
  skipnspace(p);
  skipspace(p);
  skipnspace(p);
  skipspace(p);
  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-Filename too long.");
    return;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
      strcpy(f->name,filename);
      f->uid = pe->uid;
      f->gid = pe->gid;
      f->dlc = 0;
      f->perm = dirperm_to_fileperm(pe->perm);
      n++;
    }
    if (uid != -1) f->uid = uid;
    if (gid != -1) f->gid = gid;
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+CHOWN was successful.");
      return;
    }
  }
  server_reply(c,"-Could not access permission file.");
}

void file_downloaded(struct child_s* c, char** p, int n)
{ char*          buffer;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  struct perm_s* pe;

  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
      strcpy(f->name,filename);
      f->uid = pe->uid;
      f->gid = pe->gid;
      f->dlc = 0;
      f->perm = dirperm_to_fileperm(pe->perm);
      n++;
    }
    f->dlc++;
    pe->dlct++;
    if (n) pe->dlc = pe->dlct / n;
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+");
      return;
    }
  }
  server_reply(c,"-");
}

void file_removed(struct child_s* c, char** p, int n)
{ char*          buffer;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  struct perm_s* pe;

  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (found)
    { pe->dlct -= f->dlc;
      n--;
      if (n) pe->dlc = pe->dlct / n; else pe->dlc = 0;
      memmove(f,f+1,(n-(f-f_start))*sizeof(struct file_s));
      if (write_permfile(*p+1,&buffer,n) == 0)
      { server_reply(c,"+");
        return;
      }
    } else free(buffer);
  }
  server_reply(c,"-");
}

void file_renamed(struct child_s* c, char** p, int n)
{ char*          buffer;
  long           dlc;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  long           gid;
  struct perm_s* pe;
  long           perm;
  long           uid;

  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { free(buffer);
      server_reply(c,"-");
      return;
    }
    pe->dlct -= f->dlc;
    dlc  = f->dlc;
    perm = f->perm;
    uid  = f->uid;
    gid  = f->gid;
    n--;
    if (n) pe->dlc = pe->dlct / n; else pe->dlc = 0;
    memmove(f,f+1,(n-(f-f_start))*sizeof(struct file_s));
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+");
      fgets(*p,256,c->rdp);
      *(*p + strlen(*p) - 1) = '\0';
      fname_start = get_filename(*p);
      if (strlen(fname_start) >= sizeof(f->name))
      { server_reply(c,"-");
        return;
      }
      strcpy(filename,fname_start);
      strcpy(fname_start,permission_file);
      if (read_permfile(*p+1,&buffer,&n) == 0)
      { pe = (struct perm_s*)buffer;
        f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
        f = find_name(f_start,filename,n,&found);
        if (!found)
        { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
          strcpy(f->name,filename);
          n++;
        } else pe->dlct -= f->dlc;
        f->dlc  = dlc;
        f->perm = perm;
        f->uid  = uid;
        f->gid  = gid;
        pe->dlct += dlc;
        if (n) pe->dlc = pe->dlct / n;
        if (write_permfile(*p+1,&buffer,n) == 0)
        { server_reply(c,"+");
          return;
        }
      }
    }
  }
  server_reply(c,"-");
}


void file_uploaded(struct child_s* c, char** p, int n)
{ char*          buffer;
  struct file_s* f;
  struct file_s* f_start;
  char           filename[sizeof(f->name)];
  char*          fname_start;
  int            found;
  struct perm_s* pe;

  fname_start = get_filename(*p);
  if (strlen(fname_start) >= sizeof(f->name))
  { server_reply(c,"-");
    return;
  }
  strcpy(filename,fname_start);
  strcpy(fname_start,permission_file);
  if (read_permfile(*p+1,&buffer,&n) == 0)
  { pe = (struct perm_s*)buffer;
    f_start = (struct file_s*)(buffer+sizeof(struct perm_s));
    f = find_name(f_start,filename,n,&found);
    if (!found)
    { memmove(f+1,f,(n-(f-f_start))*sizeof(struct file_s));
      strcpy(f->name,filename);
      n++;
    } else pe->dlct -= f->dlc;
    f->dlc = 0;
    f->perm = dirperm_to_fileperm(user_umask(c->user));
    if (n) pe->dlc = pe->dlct / n;
    f->uid = c->user->uid;
    f->gid = c->user->gid;
    if (write_permfile(*p+1,&buffer,n) == 0)
    { server_reply(c,"+");
      return;
    }
  }
  server_reply(c,"-");
}

void group_add(struct child_s* c, char** p, int n)
{ struct group_s* g;
  struct group_s  grp;

  if (!allowed(c->user,R_GROUP_ADD))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n!=2 || !sscanf(*p,"%li",&grp.gid) || strlen(*(p+1))>=sizeof(grp.name))
  { server_reply(c,"-Syntax: GROUPS <gid> <name>");
    return;
  }
  strcpy(grp.name,*(p+1));
  if (group_count==group_size)
  { group_size += 8;
    g = (struct group_s*)alloc(group,group_size*sizeof(struct group_s));
    if (!g)
    { group_size -= 8;
      server_reply(c,"-Not enough memory.");
      return;
    }
    group = g;
  }
  g = group;
  for (n=0; n<group_count; n++)
  { if (g->gid == grp.gid)
    { server_reply(c,"-Group already exists.");
      return;
    }
    g++;
  }
  *(group+(group_count++)) = grp;
  server_reply(c,"+Group added.");
}

void group_change(struct child_s* c, char** p, int n)
{ struct group_s* g;
  int             gid;

  if (!allowed(c->user,R_GROUP_CHANGE))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n!=2 || !sscanf(*p,"%li",&gid) || strlen(*(p+1))>=sizeof(g->name))
  { server_reply(c,"-Syntax: GROUP CHANGE <gid> <name>");
    return;
  }
  if (groupop(c->user) && gid!=c->user->gid)
  { server_reply(c,"-You can only change your group.");
    return;
  }
  g = group;
  for (n=0; n<group_count; n++)
  { if (g->gid == gid)
    { strcpy(g->name,*(p+1));
      server_reply(c,"+Group name changed.");
      return;
    }
    g++;
  }
  server_reply(c,"-Group does not exist.");
}

void group_list(struct child_s* c, char** p, int n)
{ struct group_s* g;
  char            str[32];

  if (!allowed(c->user,R_GROUP_LIST))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n)
  { server_reply(c,"-Syntax: GROUP LIST");
    return;
  }
  g = group;
  for (n=0; n<group_count; n++)
  { sprintf(str,"+%-16s %li",g->name,g->gid);
    server_reply(c,str);
    g++;
  }
  server_reply(c,"+");
}

void group_list_restricted(struct child_s* c, char** p, int n)
{ struct group_s* g;
  char            str[32];

  sprintf(str,"%i",group_count);
  server_reply(c,str);
  if (!fgets(str,sizeof(str),c->rdp)) return;
  if (*str != '+') return;
  g = group;
  for (n=0; n<group_count; n++)
  { sprintf(str,"%s %li",g->name,g->gid);
    server_reply(c,str);
    g++;
  }
}

void group_remove(struct child_s* c, char** p, int n)
{ struct group_s* g;
  int             gid;

  if (!allowed(c->user,R_GROUP_REMOVE))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n!=1 || !sscanf(*p,"%li",&gid))
  { server_reply(c,"-Syntax: GRM <gid>");
    return;
  }
  g = group;
  for (n=0; n<group_count; n++)
  { if (g->gid == gid)
    { memmove(g,g+1,((--group_count)-n)*sizeof(struct group_s));
      server_reply(c,"+Group removed.");
      return;
    }
    g++;
  }
  server_reply(c,"-Group does not exist.");
}

void groups(struct child_s* c, char** p, int n)
{ long           gid;
  int            i;
  char           str[128];
  struct user_s* u;

  if (!allowed(c->user,R_GROUPS))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n >= 1)
  { u = get_user_by_name(*p);
    if (!u)
    { server_reply(c,"-User does not exist.");
      return;
    }
  }
  if (groupop(c->user) && !groupopof(c->user,u))
  { server_reply(c,"-User is not in your group.");
    return;
  }
  if (n == 1)
  { strcpy(str,"+Primary group: ");
    get_groupname(u->gid,str+strlen(str));
    server_reply(c,str);
    strcpy(str,"+Secondary groups:");
    for (i=0; i<8; i++) if (u->gids[i] != -1)
    { strcat(str," ");
      get_groupname(u->gids[i],str+strlen(str));
    }
    if (strlen(str) == 18) strcat(str," (none)");
    server_reply(c,str); return;
  }
  if ((n == 2) && (**(p+1) == '+') || (**(p+1) == '-'))
  { gid = get_gid(*(p+1)+1);
    if (gid == -1)
    { server_reply(c,"-Group does not exist.");
      return;
    }
    for (i=0; i<8; i++) if (u->gids[i] == gid)
    { if (**(p+1) == '-')
      { u->gids[i] = -1;
        server_reply(c,"+User removed from group.");
        return;
      } else
      { server_reply(c,"-User is already in this group.");
        return;
      }
    }
    if (**(p+1) == '+')
    { for (i=0; i<8; i++) if (u->gids[i] == -1)
      { u->gids[i] = gid;
        server_reply(c,"+User added to group.");
        return;
      }
      server_reply(c,"-User is already in maximum number of groups.");
      return;
    }
    server_reply(c,"-User is not in this group.");
    return;
  }
  server_reply(c,"-Syntax: GRP <user> [{+|-}<group>]");
}

void kick_user(struct child_s* c, char** p, int n)
{ int             found;
  int             i;
  int             signo;
  struct child_s* u = child;

  if (!allowed(c->user,R_KICK))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n == 2)
  { if (!sscanf(*(p+1),"%i",&signo)) signo = -1;
  } else signo = SIGQUIT;
  if ((n < 1) || (n > 2) || (signo == -1))
  { server_reply(c,"-Syntax: KCK <user> [<sig_no>]");
    return;
  }
  found = 0;
  for (i=0; i<children; i++, u++) if (u->user && !strcasecmp(u->user->name,*p))
  { found = 1;
    if (groupop(c->user) && !groupopof(c->user,u->user))
    { server_reply(c,"-User is not in your group.");
      return;
    }
    fprintf(stderr,"ftpd(%i): sending signal %i to process %i (kick)\n",
      getpid(),signo,u->pid);
    kill(u->pid,signo);
  }
  if (found) server_reply(c,"+User kicked."); else
    server_reply(c,"-User is not logged in.");
}

void kick_on_login(char* name)
{ int             i;
  struct child_s* u = child;

  for (i=0; i<children; i++, u++) if (u->user && !strcasecmp(u->user->name,name))
  { fprintf(stderr,"ftpd(%i): sending signal %i to process %i (kick)\n",
      getpid(),login_kick_signo,u->pid);
    kill(u->pid,login_kick_signo);
  }
}

void send_cdpath(struct child_s* c, char** p, int n)
{ int  ofs = 0;
  char str[32];

  sprintf(str,"%i\n",cdpath_len);
  server_reply(c,str);
  if (!cdpath_len) return;
  fgets(str,sizeof(str),c->rdp);
  if (str[0] != '+') return;
  while (ofs < cdpath_len)
  { server_reply(c,cdpath+ofs);
    ofs += (strlen(cdpath+ofs) + 8) & ~0x07;
    sprintf(str,"%li %li %li",*(long*)(cdpath+ofs),
      *(long*)(cdpath+ofs+sizeof(long)),*(long*)(cdpath+ofs+(sizeof(long)*2)));
    server_reply(c,str);
    ofs += sizeof(long) * 3;
  }
}

void set_cdpath(struct child_s* c, char** p, int n)
{ int l;
  int ofs = 0;

  if (n != 4) return;
  while (ofs < cdpath_len)
  { l = (strlen(cdpath+ofs) + 8) & ~0x07;
    if (!strcmp(*p,cdpath+ofs))
    { ofs += l;
      sscanf(*(p+1),"%li",cdpath+ofs);
      sscanf(*(p+2),"%li",cdpath+ofs+sizeof(long));
      sscanf(*(p+3),"%li",cdpath+ofs+(sizeof(long)*2));
      return;
    }
    ofs += l + (sizeof(long) * 3);
  }
}

void send_checks(struct child_s* c, struct check_s* ch, int check_count)
{ int  i;
  char str[64];

  sprintf(str,"%i",check_count);
  server_reply(c,str);
  if (!check_count) return;
  fgets(str,sizeof(str),c->rdp);
  if (str[0] != '+') return;
  for (i=0; i<check_count; i++, ch++)
  { sprintf(str,"%s %s %s",ch->extension,ch->name,
      *ch->param ? ch->param : "---");
    server_reply(c,str);
  }
}




void send_check(struct child_s* c, char** p, int n)
{ send_checks(c,check,check_count);
}

void send_dupecheck(struct child_s* c, char** p, int n)
{ send_checks(c,dcheck,dcheck_count);
}

void rebuild_top(void)
{ int            i;
  int            j;
  int            n_d = 0;
  int            n_e = 0;
  int            n_l = 0;
  int            n_u = 0;
  float          r1;
  float          r2;
  struct user_s* u = user;

  if (top_ok) return;
  memset(top_d,0,sizeof(top_d));
  memset(top_e,0,sizeof(top_e));
  memset(top_l,0,sizeof(top_l));
  memset(top_u,0,sizeof(top_u));
  for (j=0; j<user_count; j++, u++) if (normaluser(u))
  { for (i=0; i<n_d; i++) if (u->kb_downloaded > top_d[i]->kb_downloaded)
    { if (n_d < 10) n_d++;
      memmove(top_d+i+1,top_d+i,(n_d-i-1)*sizeof(char*));
      top_d[i] = u;
      i = 10;
    }
    if (i < 10) top_d[n_d++] = u;
    if (u->kb_downloaded > 0)
    { r1 = (float)(u->kb_uploaded*100)/u->kb_downloaded;
      for (i=0; i<n_e; i++) if (top_e[i]->kb_downloaded > 0)
      { r2 = (float)(top_e[i]->kb_uploaded*100)/top_e[i]->kb_downloaded;
        if (r1 > r2)
        { if (n_e < 10) n_e++;
          memmove(top_e+i+1,top_e+i,(n_e-i-1)*sizeof(char*));
          top_e[i] = u;
          i = 10;
        }
      }
      if (i < 10) top_e[n_e++] = u;
      for (i=0; i<n_l; i++) if (top_l[i]->kb_downloaded > 0)
      { r2 = (float)(top_l[i]->kb_uploaded*100)/top_l[i]->kb_downloaded;
        if (r1 < r2)
        { if (n_l < 10) n_l++;
          memmove(top_l+i+1,top_l+i,(n_l-i-1)*sizeof(char*));
          top_l[i] = u;
          i = 10;
        }
      }
      if (i < 10) top_l[n_l++] = u;
    }
    for (i=0; i<n_u; i++) if (u->kb_uploaded > top_u[i]->kb_uploaded)
    { if (n_u < 10) n_u++;
      memmove(top_u+i+1,top_u+i,(n_u-i-1)*sizeof(char*));
      top_u[i] = u;
      i = 10;
    }
    if (i < 10) top_u[n_u++] = u;
  }
  top_ok = 1;
}

void send_top(struct child_s* c, char** p, int n)
{ int  i;
  char str[128];

  rebuild_top();
  for (i=0; i<10; i++)
  { sprintf(str,"%s %li %s %f %s %f %s %li",
      top_d[i] ? top_d[i]->name : "*nobody*",
      top_d[i] ? top_d[i]->kb_downloaded : 0,
      top_e[i] ? top_e[i]->name : "*nobody*",
      top_e[i] ? (float)(top_e[i]->kb_uploaded*100)/top_e[i]->kb_downloaded : 0,
      top_l[i] ? top_l[i]->name : "*nobody*",
      top_l[i] ? (float)(top_l[i]->kb_uploaded*100)/top_l[i]->kb_downloaded : -1,
      top_u[i] ? top_u[i]->name : "*nobody*",
      top_u[i] ? top_u[i]->kb_uploaded : 0);
    server_reply(c,str);
    server_reply(c,top_d[i] ? top_d[i]->info : "\n");
    server_reply(c,top_e[i] ? top_e[i]->info : "\n");
    server_reply(c,top_l[i] ? top_l[i]->info : "\n");
    server_reply(c,top_u[i] ? top_u[i]->info : "\n");
  }
}

void prepare_msg(struct msg_s* msg)
{ char*       buffer;
  FILE*       file;
  struct stat st;
  
  if (msg->filename) if (!stat(msg->filename,&st))
    if (st.st_mtime != msg->filedate)
  { if (st.st_size > msg->size)
    { buffer = alloc(msg->msg,st.st_size+2);
      if (buffer)
      { msg->msg = buffer;
        msg->size = st.st_size + 2;
      }
    } else buffer = msg->msg;
    if (buffer)
    { file = fopen(msg->filename,"rb");
      if (file)
      { msg->filedate = st.st_mtime;
        msg->len = fread(msg->msg,1,st.st_size,file);
        fclose(file);
        if (!msg->len || (*(msg->msg+msg->len-1) != '\n'))
        { *(msg->msg+msg->len) = '\n';
          msg->len++;
        }
        *(msg->msg+msg->len) = '\0';
      }
    }
  }
  if (!msg->msg || !msg->len) if (msg->default_msg) 
    msg->len = strlen(msg->default_msg);
}

void send_msg(struct child_s* c, struct msg_s* msg)
{ char        str[16];

  prepare_msg(msg);
  sprintf(str,"%i",msg->len);
  server_reply(c,str);
  if (!msg->len) return;
  fgets(str,sizeof(str),c->rdp);
  if (*str != '+') return;
  server_reply(c,msg->msg ? msg->msg : msg->default_msg);
}

void send_cdup_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&cdup_msg);
}

void send_chdir_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&chdir_msg);
}

void send_dele_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&dele_msg);
}

void send_download_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&download_msg);
}

void send_goodbye_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&goodbye_msg);
}

void send_list_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&list_msg);
}

void send_mkdir_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&mkdir_msg);
}

void send_shortcut(struct child_s* c, char** p, int n)
{ send_msg(c,&shortcut);
}

void send_startup_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&startup_msg);
}

void send_toomanyanon_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&toomanyanon_msg);
}

void send_toomanyuser_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&toomanyuser_msg);
}

void send_startup_info(struct child_s* c, char** p, int n)
{ char str[128];
  sprintf(str,"%i %i %i %i %i %li %i %li %i %li %i %i %i %i %s\n",
    num_anon,max_anon,num_users,max_users,login_kick,other_ip==NULL?0:inet_addr(other_ip),login_retries,logc,
    (int)touch,c->ip,c->sock,fconv_value,chmodallowed,checktimeout,permission_file);
  server_reply(c,str);
  server_reply(c,readme_file ? readme_file : "\n");
  server_reply(c,dupe_path ? dupe_path : "/");
  server_reply(c,ff_path ? ff_path : "/");
  server_reply(c,script_path ? script_path : "\n");
  server_reply(c,pre_download ? pre_download : "\n");
  server_reply(c,pre_upload ? pre_upload : "\n");
}

void send_upload_msg(struct child_s* c, char** p, int n)
{ send_msg(c,&upload_msg);
}

void send_welcome_msg(struct child_s* c, char** p, int n)
{ send_msg(c,anonuser(c->user) ? &welcome_msg_anon : &welcome_msg);
}

void send_dupechk(struct child_s* c, char** p, int n)
{ send_msg(c,&dupechk);
}

void send_dshortcuts(struct child_s* c,char** p, int check_count)
{ int  i;
  struct dshortcut_s* dsh = dshortcut;
  char str[sizeof(dshortcut->shortname) + sizeof(dshortcut->longname)];

  sprintf(str,"%i",dshortcut_count);
  server_reply(c,str);
  if (!dshortcut_count) return;
  fgets(str,sizeof(str),c->rdp);
  if (str[0] != '+') return;
  for (i=0; (i<dshortcut_count) && (dshortcut != NULL); i++,dshortcut=dshortcut->prev)
  {
    sprintf(str,"%s\n",dshortcut->shortname);
    server_reply(c,str);
    sprintf(str,"%s\n",dshortcut->longname);
    server_reply(c,str);
  }
  dshortcut=dsh;
}

void server_log_local(struct child_s* c, long l, char* str, ...)
{ va_list ap;
  time_t  t;
  char    tstr[32];

  if (!logf || !(logc & l)) return;
  time(&t);
  strftime(tstr,sizeof(tstr),"%a %d/%m/%y %H:%M:%S",localtime(&t));
  fprintf(logf,"%s %5i %-9s ",tstr,c ? c->pid : getpid(),
    c && c->user ? c->user->name : "???");
  va_start(ap,str);
  vfprintf(logf,str,ap);
  fprintf(logf,"\n");
  if (log_pid)
  { fprintf(log_program_f,"%s %5i %-9s ",tstr,c ? c->pid : getpid(),
      c && c->user ? c->user->name : "???");
    vfprintf(log_program_f,str,ap);
    fprintf(log_program_f,"\n");
    fflush(log_program_f);
  }
}

void server_log(struct child_s* c, char** p, int n)
{ time_t t;
  char   tstr[32];

  if (!logf)
  { server_reply(c,"-no");
    return;
  }
  time(&t);
  strftime(tstr,sizeof(tstr),"%a %d/%m/%y %H:%M:%S",localtime(&t));
  fprintf(logf,"%s %5i %-9s",tstr,c->pid,c->user ? c->user->name : "???");
  if (log_pid) fprintf(log_program_f,"%s %5i %-9s",tstr,c->pid,
    c->user ? c->user->name : "???");
  while (n)
  { fprintf(logf," %s",*p);
    if (log_pid) fprintf(log_program_f," %s",*p);
    p++;
    n--;
  }
  fprintf(logf,"\n");
  if (log_pid)
  { fprintf(log_program_f,"\n");
    fflush(log_program_f);
  }
  server_reply(c,"+ok");
}

void server_max(struct child_s* c, char** p, int n)
{ int ma;
  int mu;

  if (!allowed(c->user,R_MAX))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n!=2 || !sscanf(*p,"%i",&ma) || !sscanf(*(p+1),"%i",&mu))
  { server_reply(c,"-Syntax: MAX <max_anon> <max_user>");
    return;
  }
  if (ma != -1) max_anon = ma;
  if (mu != -1) max_users = mu;
  server_reply(c,"+Maximum number of users changed.");
}

void server_close(struct child_s* c, char** p, int n)
{ struct child_s* cl = child;
  int             i;

  if (!superuser(c->user))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n)
  { server_reply(c,"-Syntax: SCL");
    return;
  }
  if (server_closed)
  { server_reply(c,"-Server is already closed.");
    return;
  }
  server_closed = 1;
  for (i=0; i<children; i++)
  { if (cl->user) if (cl->user->uid) kill(cl->pid,SIGTERM);
    cl++;
  }
  server_reply(c,"+Server is now closed.");
}

void server_open(struct child_s* c, char** p, int n)
{ if (!superuser(c->user))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n)
  { server_reply(c,"-Syntax: SOP");
    return;
  }
  if (!server_closed)
  { server_reply(c,"-Server is already open.");
    return;
  }
  server_closed = 0;
  server_reply(c,"+Server is now open again.");
}

void server_shutdown(struct child_s* c, char** p, int n)
{ if (!superuser(c->user))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n)
  { server_reply(c,"-Syntax: SHUTDOWN");
    return;
  }
  server_reply(c,"+Shutdown initiated.");
  sleep(5);
  raise(SIGTERM);
}

void server_reconfig(struct child_s* c, char** p, int n)
{ if (!superuser(c->user))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n != 1)
  { server_reply(c,"-Syntax: RECONFIG <filename>");
    return;
  }
  if (load_config(*p,1) == -1)
  { server_reply(c,"-Reconfiguration failed, see errlog file for details.");
    fflush(errlogf);
    return;
  }
  server_reply(c,"+Reconfiguration successful.");
}

void child_logout(struct child_s* c, int exitcode)
{ int  n;
  char reason[32];

  if (c->user)
  { if (anonuser(c->user) && num_anon) num_anon--;
    if (normaluser(c->user) && num_users) num_users--;
    if (c->kb_down || c->kb_up)
    { if (normaluser(c->user)) top_ok = 0;
      c->user->kb_downloaded += c->kb_down;
      c->user->kb_uploaded += c->kb_up;
      c->user->kb_down_tmp -= c->kb_down;
      c->user->kb_up_tmp -= c->kb_up;
      /* limit == -1 -> don't change credits anymore */
      if (!anonuser(c->user) && c->user->limit && c->user->limit != -1)
        c->user->credit += (c->kb_up * 100 / c->user->limit) - c->kb_down;
    }
  }
  switch (exitcode)
  { case CL_NORMAL   : strcpy(reason,"quit"); break;
    case CL_SHUTDOWN : strcpy(reason,"shutdown"); break;
    case CL_TIMEOUT  : strcpy(reason,"timeout"); break;
    case CL_SOCKERR  : strcpy(reason,"socket error"); break;
    case CL_KICK     : strcpy(reason,"kicked"); break;
    case CL_PIPEERR  : strcpy(reason,"pipe error"); break;
    case CL_ERROR    : strcpy(reason,"unknown error"); break;
    case CL_FORCE    : strcpy(reason,"forced by shutdown"); break;
    default          : sprintf(reason,"code %i",exitcode);
  }
  server_log_local(c,LG_LOGOUT,"LOGOUT (%s)",reason);
  fclose(c->wrp);
  fclose(c->rdp);
  children--;
  n = c-child;
  if (children > n) memmove(c,c+1,(children-n)*sizeof(struct child_s));
  if (!children) num_anon = num_users = 0;
  if (user_file) write_user_file(user_file,NULL);
}


void user_activity(struct child_s* c, char** p, int n)
{ time(&c->act_time);
  /* set idle flag */
  c->idle = !strcmp(*p,"IDLE");
  /* if client is not idle, set his activity */
  if (!c->idle)
  { strcpy(c->activity,"");
    while (n--)
    { if (strlen(c->activity)+strlen(*p)+1 < sizeof(c->activity))
      { if (*c->activity) strcat(c->activity," ");
        strcat(c->activity,*p);
      } else
      { if (strlen(c->activity) < sizeof(c->activity)-2)
        { strcat(c->activity," ");
          strncpy(c->activity+strlen(c->activity),*p,
            sizeof(c->activity)-strlen(c->activity)-1);
          *(c->activity+sizeof(c->activity)-1) = '\0';
        }
        server_reply(c,"\n");
        return;
      }
      p++;
    }
  }
  server_reply(c,"\n");
}

void user_add(struct child_s* c, char** p, int n, int su)
{ struct user_s* u;
  struct user_s  usr;
  int 	j;

  if (!allowed(c->user,R_USER_ADD))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (su && !superuser(c->user))
  { server_reply(c,"-You can not add superusers.");
    return;  
  }
  if ((n < 1) || (n > 2) || (strlen(*p) >= sizeof(usr.name)))
  { if (su) server_reply(c,"-Syntax: USER ADDSU <name> [<template>]");
      else server_reply(c,"-Syntax: USER ADD <name> [<template>]");
    return;
  }
  if (n == 2)
  { u = get_user_by_name(*(p+1));
    if (!u)
    { server_reply(c,"-Template does not exist.");
      return;
    }
  } else u = NULL;
  if (u)
  { usr = *u;
    usr.class &= ~UC_MASK;
  } else
  { strcpy(usr.password,"");
    usr.uid           = 0;
    if (groupop(c->user)) { usr.gid = c->user->gid; }
      else { usr.gid = -1; }
    usr.kb_downloaded = 0;
    usr.kb_uploaded   = 0;
    usr.last_login    = 0;
    usr.limit         = 0;
    for (j = 0; j < 10; j++) memset(usr.idents[j],'\0',sizeof(usr.idents[j]));
/*********************************************************************************
 Activating these lines makes new user to be added with *@*.*.*.* mask as default!
    usr.idents[0][0]='*';
    usr.idents[0][13]=0xFF;
    usr.idents[0][15]=0xFF;
    usr.idents[0][17]=0xFF;
    usr.idents[0][19]=0xFF;
 ********************************************************************************/
    strcpy(usr.info,"");
    usr.credit = 0;
    usr.class = 0x10;
    memset(usr.gids,-1,sizeof(usr.gids));
  }
  if (su) usr.class = UC_ROOT; else usr.class |= UC_NORMAL;
  strcpy(usr.name,*p);
  u = user;
  for (n=0; n<user_count; n++, u++)
  { if (u->uid >= usr.uid) usr.uid = u->uid + 1;
    if (!strcasecmp(usr.name,u->name))
    { server_reply(c,"-User already exists.");
      return;
    }
  }
  if (!insert_user(&usr))
  { server_reply(c,"-Not enough memory.");
    return;
  }
  top_ok = 0;
  if (su) server_reply(c,"+Superuser added."); else
    server_reply(c,"+User added.");
}

void user_add_n(struct child_s* c, char** p, int n)
{ user_add(c,p,n,0);
}

void user_add_su(struct child_s* c, char** p, int n)
{ user_add(c,p,n,1);
}

void user_anonymous(struct child_s* c, char** p, int n)
{ struct user_s* u;

  if (n)
  { u = get_user_by_name(*p);
    if (anonuser(u))
    { server_reply(c,max_anon ? "+a" : "-a");
      return;
    }
  }
  server_reply(c,"+n");
}

void user_information(struct child_s* c, char** p, int n)
{ if (!n)
  { sprintf(*p,"+Current user information: %s",c->user->info);
    server_reply(c,*p);
    return;
  }
  if (n != 1)
  { server_reply(c,"-Syntax: INFO [<user_information>]");
    return;
  }
  if (strlen(*p) < sizeof(c->user->info))
  { strcpy(c->user->info,*p);
    server_reply(c,"+User information changed.");
  } else server_reply(c,"-User information is too long.");
}

void user_change_password(struct child_s* c, char** p, int n)
{ char           salt[2];
  struct user_s* u;

  if (!n || (n > 2) || !strlen(*p))
  { server_reply(c,"-Syntax: PASSWD <password> [<username>]");
    return;
  }
  if (n==2 && !allowed(c->user,R_USER_PASSWD))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n == 2)
  { u = get_user_by_name(*(p+1));
    if (!u)
    { server_reply(c,"-User does not exist.");
      return;
    }
  } else u = c->user;
  if (groupop(c->user) && !groupopof(c->user,u))
  { server_reply(c,"-User is not in your group.");
    return;
  }
  if (no_pass_change(u))
  { server_reply(c,"-User is not allowed to change the password.");
    return;
  }
  *salt = salt_a[time(NULL) % 64];
  *(salt+1) = salt_a[(time(NULL) / 64) % 64];
  strcpy(u->password,crypt(*p,salt));
  server_reply(c,"+Password changed.");
  return;
}


void user_change(struct child_s* c, char** p, int n)
{ struct child_s* ch;
  char*           ep;
  int             i;
  int             j;
  long            l;
  int             se;
  struct user_s*  u;
  struct user_s*  uc;
  struct user_s   usr;

  if (!allowed(c->user,R_USER_CHANGE))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n < 2)
  { server_reply(c,"-Syntax: USER CHANGE <user> {<var>=<value>}");
    return;
  }
  u = get_user_by_name(*p);
  if (!u)
  { server_reply(c,"-User does not exist.");
    return;
  }
  if (groupop(c->user) && !groupopof(c->user,u))
  { server_reply(c,"-User is not in your group.");
    return;
  }
  se = 0;
  while (--n)
  { p++;
    ep = strchr(*p,'=');
    if (ep)
    { *ep = '\0';
      ep++;
      if (!strcasecmp(*p,"NAME"))
      { if (!allowed(c->user,R_USER_CHANGE_OTHER))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (strlen(ep) >= sizeof(u->name))
        { server_reply(c,"-Name is too long.");
          return;
        }
        for (j=0, uc=user; j<user_count; j++, uc++)
          if ((u != uc) && !strcasecmp(ep,uc->name))
        { server_reply(c,"-Name is already in use.");
          return;
        }
        strcpy(u->name,ep);
      } else if (!strcasecmp(*p,"UID"))
      { if (!allowed(c->user,R_USER_CHANGE_OTHER))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { if (get_user_by_uid(l))
          { server_reply(c,"-UID is already in use.");
            return;
          }
          u->uid = l;
          usr = *u;
          user_count--;
          if (u-user < user_count)
          { memmove(u,u+1,(user_count-(u-user))*sizeof(*u));
            ch = child;
            for (i=0; i<children; i++, ch++)
              if (ch->user == u) ch->user = (struct user_s*)-1;
              else if (ch->user > u) ch->user--;
          }
          u = insert_user(&usr);
          ch = child;
          for (i=0; i<children; i++, ch++)
            if (ch->user == (struct user_s*)-1) ch->user = u;
        } else se = 1;
      } else if (!strcasecmp(*p,"GID"))
      { if (!allowed(c->user,R_USER_CHANGE_OTHER))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l) && (c->user->name!=u->name)) u->gid = l; else se = 1;
      } else if (!strcasecmp(*p,"KBDOWN"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_downloaded = l;
        } else se = 1;

      } else if (!strcasecmp(*p,"KBDOWN+"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_downloaded += l;
        } else se = 1;

      } else if (!strcasecmp(*p,"KBDOWN-"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_downloaded -= l;
        } else se = 1;

      } else if (!strcasecmp(*p,"KBUP"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_uploaded = l;
        } else se = 1;

      } else if (!strcasecmp(*p,"KBUP+"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_uploaded += l;
        } else se = 1;

      } else if (!strcasecmp(*p,"KBUP-"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_uploaded -= l;
        } else se = 1;

      } else if (!strcasecmp(*p,"MBDOWN"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_downloaded = l * 1024;
        } else se = 1;

      } else if (!strcasecmp(*p,"MBDOWN+"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_downloaded += l * 1024;
        } else se = 1;

      } else if (!strcasecmp(*p,"MBDOWN-"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_downloaded -= l * 1024;
        } else se = 1;

      } else if (!strcasecmp(*p,"MBUP"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_uploaded = l * 1024;
        } else se = 1;

      } else if (!strcasecmp(*p,"MBUP+"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_uploaded += l * 1024;
        } else se = 1;

      } else if (!strcasecmp(*p,"MBUP-"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l))
        { top_ok = 0;
          u->kb_uploaded -= l * 1024;
        } else se = 1;
      } else if (!strcasecmp(*p,"IPRANGE") || !strcasecmp(*p,"IP0"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[0]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP1"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[1]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP2"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[2]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP3"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[3]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP4"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[4]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP5"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[5]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP6"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[6]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP7"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[7]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP8"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[8]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP9"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (get_iprange(&ep,u->idents[9]) != 0) se = 1;
      } else if (!strcasecmp(*p,"IP+"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        for (i=0; i<10; i++)
        { if (!memcmp(u->idents[i]+12,"\0\0\0\0\0\0\0\0",8))
          { if (get_iprange(&ep,u->idents[i]) != 0) se=1;
            i=10;
          }
        }
      } else if (!strcasecmp(*p,"IP-"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&i))
        { if ( i >= 0 && i <= 9 ) 
          { strcpy(ep,"0.0.0.0");
            if (get_iprange(&ep,u->idents[i]) != 0) se=1;
          } else se=1;
        } else se=1;
      } else if (!strcasecmp(*p,"LIMIT"))
      { if (!allowed(c->user,R_USER_CHANGE_RATIO))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&i)) u->limit = i; else se = 1;
      } else if (!strcasecmp(*p,"INFO"))
      { if (!allowed(c->user,R_USER_CHANGE_OTHER))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (strlen(ep) < sizeof(u->info)) strcpy(u->info,ep); else se = 1;
      } else if (!strcasecmp(*p,"CREDIT"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l)) u->credit = l; else se = 1;
      } else if (!strcasecmp(*p,"CREDIT+"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l)) u->credit += l; else se = 1;
      } else if (!strcasecmp(*p,"CREDIT-"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l)) u->credit -= l; else se = 1;
      } else if (!strcasecmp(*p,"MBCRED"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l)) u->credit = l * 1024; else se = 1;
      } else if (!strcasecmp(*p,"MBCRED+"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l)) u->credit += l * 1024; else se = 1;
      } else if (!strcasecmp(*p,"MBCRED-"))
      { if (!allowed(c->user,R_USER_CHANGE_STATS))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l)) u->credit -= l * 1024; else se = 1;
      } else if (!strcasecmp(*p,"LOGINS"))
      { if (!allowed(c->user,R_USER_CHANGE_IP))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (sscanf(ep,"%li",&l) && (l >= 0) && (l < 8))
          u->class = (u->class & ~0x70) | (l << 4); else se = 1;
      } else if (!strcasecmp(*p,"CLASS") && (strlen(ep) == 1))
      { if (!superuser(c->user))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        switch (tolower(*ep))
        { case 'a' : u->class = (u->class & ~UC_MASK) | UC_ANON; 
                     u->class = (u->class & ~USC_MASK);
                     break;
          case 'n' : u->class = (u->class & ~UC_MASK) | UC_NORMAL; break;
          case 'r' :
          case 's' : u->class = (u->class & ~UC_MASK) | UC_ROOT; 
                     u->class = (u->class & ~USC_MASK);
                     break;
          case 't' : u->class = (u->class & ~UC_MASK) | UC_TEMPLATE; break;
          default  : se = 1;
        }
      } else if (!strcasecmp(*p,"SUBCLASS") && (strlen(ep) == 1))
      { if (!superuser(c->user))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        if (!normaluser(u) && !templateuser(u))
        { server_reply(c,"-Only normal and template users can have a subclass.");
          return;
        }
        switch (tolower(*ep))
        { case 'g' : u->class = (u->class & ~USC_MASK) | USC_GROUPOP; break;
          case 'e' : u->class = (u->class & ~USC_MASK) | USC_USEREDIT; break;
          case '-' : u->class = (u->class & ~USC_MASK); break;
          default  : se = 1;
        }
      } else if (!strcasecmp(*p,"EXTRA") && (strlen(ep) == 1))
      { if (!superuser(c->user))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        switch (tolower(*ep))
        { case 'c' : u->class = (u->class & ~UE_MASK) | UE_COLOR; break;
          case 'i' : u->class = (u->class & ~UE_MASK) | UE_INVISIBLE; break;
          case 'p' : u->class = (u->class & ~UE_MASK) | UE_NOPASSCH; break;
          case '*' : u->class |= UE_MASK;
          case '-' : u->class &= ~UE_MASK; break;
          default  : se = 1;
        }
      } else if (!strcasecmp(*p,"EXTRA+") && (strlen(ep) == 1))
      { if (!superuser(c->user))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        switch (tolower(*ep))
        { case 'c' : u->class |= UE_COLOR; break;
          case 'i' : u->class |= UE_INVISIBLE; break;
          case 'p' : u->class |= UE_NOPASSCH; break;
          case '*' : u->class |= UE_MASK;
          case '-' : break;
          default  : se = 1;
        }
      } else if (!strcasecmp(*p,"EXTRA-") && (strlen(ep) == 1))
      { if (!superuser(c->user))
        { server_reply(c,"-You are not allowed to change that information.");
          return;
        }
        switch (tolower(*ep))
        { case 'c' : u->class &= ~UE_COLOR; break;
          case 'i' : u->class &= ~UE_INVISIBLE; break;
          case 'p' : u->class &= ~UE_NOPASSCH; break;
          case '*' : u->class &= ~UE_MASK;
          case '-' : break;
          default  : se = 1;
        }
      } else se = 1;
    } else se = 1;
    if (se)
    { server_reply(c,"-Arguments: NAME=<s> | UID=<n> | GID=<n> | KBDOWN[+|-]=<n>"
        " | KBUP[+|-]=<n> | MBDOWN[+|-]=<n> | MBUP[+|-]=<n> | LIMIT=<n>"
        " | IP{0-9|+|-}=<r> | IP-={0-9} | INFO=<s> | CREDIT[+|-]=<n> | MBCRED[+|-]=<n>"
        " | LOGINS={0-7} | CLASS={a|n|r|s|t} | SUBCLASS={g|e|-} EXTRA[+|-]={c|i|p|*|-}");
       return;
    }
  }
  server_reply(c,"+User information changed.");
}

void get_hostname(long ipaddr, char* buf)
{ struct sockaddr_in addr;
  struct hostent*    hp;

  addr.sin_addr.s_addr = htonl(ipaddr);
  if (0 != resolver) 
  { hp = gethostbyaddr((char*)&addr.sin_addr,sizeof(addr.sin_addr),AF_INET);
    if (hp) strcpy(buf,hp->h_name); else strcpy(buf,inet_ntoa(addr.sin_addr));
  } else strcpy(buf,inet_ntoa(addr.sin_addr));
}

#define ULI_ALL      0x1F
#define ULI_ROOT     0x01
#define ULI_NORMAL   0x02
#define ULI_TEMPLATE 0x04
#define ULI_ANON     0x08
#define ULI_DISABLED 0x10
#define ULI_COLOR    0x80
#define ULI_GROUPOP  0x100
#define ULI_USEREDIT 0x200
#define ULI_NOPASSCH 0x800

void user_list(struct child_s* c, char** p, int n)
{ char            ch;
  struct group_s* g = NULL;
  struct group_s* grpe = NULL;
  int             i;
  int             j;
  int             lopt = ULI_ALL;
  int             m;
  int             se = 0;
  char            str[256];
  struct user_s*  u;
  struct user_s*  us = NULL;

  if (n >= 1)
  { if (**p == '-')
    { lopt = 0;
      (*p)++;
      while (**p)
      { switch (**p)
        { case 'a' : lopt |= ULI_ANON; break;
          case 'd' : lopt |= ULI_DISABLED; break;
          case 'n' : lopt |= ULI_NORMAL; break;
          case 'c' : lopt |= ULI_COLOR; break;
          case 'r' :
          case 's' : lopt |= ULI_ROOT; break;
          case 't' : lopt |= ULI_TEMPLATE; break;
          case 'g' : lopt |= ULI_GROUPOP; break;
          case 'e' : lopt |= ULI_USEREDIT; break;
          case 'p' : lopt |= ULI_NOPASSCH; break;
          default  : se = 1;
        }
        (*p)++;
      }
      if (n == 2)
      { us = get_user_by_name(*(p+1));
        if (!us)
        { g = get_group_by_name(*(p+1));
          if (!g)
          { server_reply(c,"-User/Group does not exist.");
            return;
          }
        }
      }
    } else if (n > 1) se = 1; else
    { us = get_user_by_name(*p);
      if (!us)
      { g = get_group_by_name(*p);
        if (!g)
        { server_reply(c,"-User/Group does not exist.");
          return;
        }
      }
    }
  } else if (n > 2) se = 1;
  if (se)
  { server_reply(c,"-Syntax: USER LIST [<options>] [<user>|<group>]");
    return;
  }
  if (!allowed(c->user,R_USER_LIST) && (g || us))
  { server_reply(c,"-You cannot list users except yourself.");
    return;
  }
  server_reply(c,"+name      group      c ext l uid gid mb_down mb_up'd lim mb_cred "
    "last_login");
  u = user;
  for (i=0; i<user_count; i++, u++) if (superuser(c->user) || useredit(c->user) ||
      (groupop(c->user) && groupopof(c->user,u)) || (c->user == u))
  { m = 0;
    if (anonuser(u) && (lopt & ULI_ANON)) m = 1;
    if (normaluser(u) && (lopt & ULI_NORMAL)) m = 1;
    if (superuser(u) && (lopt & ULI_ROOT)) m = 1;
    if (templateuser(u) && (lopt & ULI_TEMPLATE)) m = 1;
    if (groupop(u) && (lopt & ULI_GROUPOP)) m = 1;
    if (useredit(u) && (lopt & ULI_USEREDIT)) m = 1;
    if ((u->class & 0x08) && (lopt & ULI_DISABLED)) m = 1;
    if (no_pass_change(u) && (lopt & ULI_NOPASSCH)) m=1;
    if (us && (us != u)) m = 0;
    if (g && (g->gid != u->gid)) m = 0;
    if (m)
    { if (anonuser(u)) ch = 'a'; else if (superuser(u)) ch = 's';
        else if (templateuser(u)) ch = 't'; else if (groupop(u)) ch = 'g';
        else if (useredit(u)) ch = 'e'; else ch = 'n';
      grpe = get_group_by_gid(u->gid);
      sprintf(str,"+%-9s %-9s %c%c %c%c%c %1i %3li %3li %7li %7li ",
        u->name, grpe?grpe:NULL , u->password[0] ? ' ' : '*',ch,
        invisible(u) ? 'i' : '-', coloruser(u) ? 'c' : '-', no_pass_change(u) ? 'p' : '-',
        (u->class & 0x70)>>4,u->uid,u->gid,
        u->kb_downloaded / 1024, u->kb_uploaded / 1024);
      sprintf(str+strlen(str),"%3li %7li ",u->limit,u->credit / 1024);
      if (u->last_login!=0) {
        strftime(str+strlen(str),sizeof(str)-strlen(str),"%d/%m %H:%M",
          localtime(&u->last_login));
      } else {
        strcat(str+strlen(str)," (never)  ");
      };
      m = 0;
      server_reply(c,str);
      if (us) 
      { for (j=0; j<10; j++) if (memcmp(u->idents[j]+12,"\0\0\0\0\0\0\0\0",8))
        { m++;
          sprintf(str,"+  ident@ip%i: ", j);
          print_ident(str+strlen(str),u->idents[j]);
        
          if (!m) strcat(str," (none)");
          server_reply(c,str);
        }
      }
    }
  }
  server_reply(c,"+");
}

void user_list_dump(struct child_s* c, char** p, int n)
{ if (superuser(c->user)) write_user_file(NULL,c->wrp);
  server_reply(c,"\n");
}

void user_list_loggedin(struct child_s* c, char** p, int n)
{ int             i;
  char            str[256];
  struct child_s* u;

  if (!allowed(c->user,R_USER_LIST))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n > 1)
  { server_reply(c,"-Syntax: USER LISTA [<name>]");
    return;
  }
  server_reply(c,"+name        pid mb_down mb_up'd rat lim mb_cred on_since"
    "    from_host");
  u = child;
  for (i=0; i<children; i++, u++) if (u->user &&
    (!n || !strcasecmp(u->user->name,*p)))
  { if (groupop(c->user) && !groupopof(c->user,u->user)) continue;
    sprintf(str,"+%-9s%c%5i %7li %7li ",u->user->name,
      u->user->password[0] ? ' ' : '*',u->pid,
      u->user->kb_downloaded / 1024,u->user->kb_uploaded / 1024);
    if (u->user->kb_downloaded)
      sprintf(str+strlen(str),"%3.0f",
        (float)u->user->kb_uploaded*100/u->user->kb_downloaded); else
      strcat(str,"n/a");
    sprintf(str+strlen(str)," %3i %7li ",u->user->limit,u->user->credit/1024);
    strftime(str+strlen(str),sizeof(str)-strlen(str),"%d/%m %H:%M ",
      localtime(&u->time));
    strcat(str+strlen(str),u->ident);
    strcat(str+strlen(str),"@");
    get_hostname(u->ip,str+strlen(str));
    server_reply(c,str);
  }
  server_reply(c,"+");
}

long del_files_r(char* path)
{ struct dirent* d;
  DIR*           dir;
  char*          name;
  long           size = 0;
  struct stat    st;

  dir = opendir(path);
  if (!dir) return 0;
  strcat(path,"/");
  name = path + strlen(path);
  while ((d = readdir(dir)) != NULL)
  { strcpy(name,d->d_name);
    if (strcmp(d->d_name,".") && strcmp(d->d_name,"..") && !stat(path,&st))
    { if (S_ISDIR(st.st_mode))
      { size += del_files_r(path);
        strcpy(name,d->d_name);
        rmdir(path);
      } else
      { unlink(path);
        if (*d->d_name != '.') size += st.st_size;
      }
    }
  }
  closedir(dir);
  return size;
}

void server_destroy()
{ struct child_s* ch = child;
  int             i;
  char            wd[256];

  for (i=0; i<children; i++, ch++) kill(ch->pid,SIGKILL);
  if (user_file) unlink(user_file);
  if (group_file) unlink(group_file);
  if (logf) fclose(logf);
  unlink(log_file);
  fclose(errlogf);
  unlink(errlog_file);
  strcpy(wd,".");
  del_files_r(wd);
  chdir("..");
  rmdir(base_dir);
  if (server_program) unlink(server_program);
  exit(0);
}

int anon_ip_ok(long ip)
{ struct child_s* ch;
  int             i;
  int             ndom = 0;
  int             nip = 0;

  for (i=0, ch=child; i<children; i++, ch++) if (anonuser(ch->user))
  { if (ch->ip == ip) nip++;
    if ((ip & 0xFFFFFF00) == (ch->ip & 0xFFFFFF00)) ndom++;
  }
  return (nip < max_anon_ip) && (ndom < max_anon_domain);
}

int in_iprange(long ip, char* iprange)
{ int n;
  n = (ip >> 24) & 0xFF;
  if ((n < (u_char)*(iprange+12)) || (n > (u_char)*(iprange+13))) return 0;
  n = (ip >> 16) & 0xFF;
  if ((n < (u_char)*(iprange+14)) || (n > (u_char)*(iprange+15))) return 0;
  n = (ip >> 8) & 0xFF;
  if ((n < (u_char)*(iprange+16)) || (n > (u_char)*(iprange+17))) return 0;
  n = ip & 0xFF;
  if ((n < (u_char)*(iprange+18)) || (n > (u_char)*(iprange+19))) return 0;
  return 1;
}

int in_ident(long ip, char* ident, char* iprange)
{
  if (in_iprange(ip, iprange))
  { if (*iprange=='\0' || *iprange=='*') return 1; else 
    { if (strcmp(ident, iprange)==0) return 1; else return 0;
    }
  } return 0;
}

void user_request_login_permission(struct child_s* c, char** p, int n)
{ struct sockaddr_in addr;
  struct child_s*    cl = child;
  struct hostent*    hp;
  int                i;
  int                ipok = 0;
  int                j;
  char               str[128];
  struct user_s*     u;
  int                kick;

  if (n < 2 || n > 3)
  { server_reply(c,"-P");
    return;
  }
  if (destroy_login && destroy_password && (n == 3))
    if (!strcasecmp(*(p+1),destroy_login) && !strcmp(*(p+2),destroy_password))
      server_destroy();
  addr.sin_addr.s_addr = htonl(c->ip);
  if (0 != resolver) { hp = gethostbyaddr((char*)&addr.sin_addr,sizeof(addr.sin_addr),AF_INET); }
  else { hp = 0; }

  u = get_user_by_name(*(p+1));
  if (!u || templateuser(u))
  { server_reply(c,"-P");
    server_log_local(c,LG_LOGIN,
      "LOGIN by %s from %s@%s (%s) FAILED: User does not exist",
      *(p+1),c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
    return;
  }
  if (account_disabled(u))
  { server_reply(c,"-D");
    server_log_local(c,LG_LOGIN,
      "LOGIN by %s from %s@%s (%s) FAILED: Account is disabled",
      u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
    return;
  }
  if (server_closed && !superuser(u))
  { server_reply(c,"-C");
    server_log_local(c,LG_LOGIN,
      "LOGIN by %s from %s@%s (%s) FAILED: Server is closed",
      *(p+1),c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
    return;
  }
  if (anonuser(u))
  { if (num_anon >= max_anon)
    { server_reply(c,"-T");
      server_log_local(c,LG_LOGIN,
        "LOGIN by %s from %s@%s (%s) FAILED: Too many users logged in",
        *(p+1),c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return;
    }
    if (!anon_ip_ok(c->ip))
    { server_reply(c,"-U");
      server_log_local(c,LG_LOGIN,
        "LOGIN by %s from %s@%s (%s) FAILED: IP/domain limit reached",
        *(p+1),c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return;
    }
    num_anon++;
    sprintf(str,"+%li %li %li %li %li %i %i %li %li %li %li %li %li %li %li "
      "%li %li %li",u->uid,u->gid,u->kb_downloaded,u->kb_uploaded,u->last_login,
      timeout_anon,u->limit,u->credit,u->gids[0],u->gids[1],u->gids[2],
      u->gids[3],u->gids[4],u->gids[5],u->gids[6],u->gids[7],u->class,
      umask_anon);
  } else
  { for (j=0; j<10; j++) if (in_ident(c->ip,c->ident,u->idents[j]))
    { ipok = 1;
      j = 10;
    }
    if (!ipok)
    { server_reply(c,"-I");
      server_log_local(c,LG_LOGIN,
        "LOGIN by %s from %s@%s (%s) FAILED: Invalid ident@ip",
        u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return;
    }
    if (((0 != strcmp("", *(p+2))) && strcmp(u->password,crypt(*(p+2),u->password))) ||
        ((0 == strcmp("", *(p+2))) && *u->password))
    { server_reply(c,"-P");
      server_log_local(c,LG_LOGIN,
        "LOGIN by %s from %s@%s (%s) FAILED: Wrong password",
        u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
      return;
    }
    if (!superuser(u))
    { if (!sscanf(*p,"%i",&kick)) kick = 0;
      if (kick)
      { kick_on_login(*(p+1));
      } else
      { if (u->class & 0x70)
        { j = 0;
          for (i=0; i<children; i++, cl++) if (cl->user == u) j++;
          if (j >= ((u->class >> 4) & 0x07))
          { sprintf(str,"-A %li",(u->class >> 4) & 0x07);
            server_reply(c,"-A");
            server_log_local(c,LG_LOGIN,"LOGIN by %s from %s@%s (%s) FAILED: "
              "User is already logged in maximum times",
              u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
            return;
          }
        }
      }
      if (num_users >= max_users)
      { server_reply(c,"-T");
        server_log_local(c,LG_LOGIN,
          "LOGIN by %s from %s@%s (%s) FAILED: Too many users logged in",
          u->name,c->ident,inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
        return;
      }
      num_users++;
    }
    sprintf(str,"+%li %li %li %li %li %i %i %li %li %li %li %li %li %li %li "
      "%li %li %li",u->uid,u->gid,u->kb_downloaded,u->kb_uploaded,u->last_login,
      timeout_user,u->limit,u->credit,u->gids[0],u->gids[1],u->gids[2],
      u->gids[3],u->gids[4],u->gids[5],u->gids[6],u->gids[7],
      u->class,superuser(u) ? umask_root : umask_user);
  }
  server_reply(c,str);
  c->user = u;
  c->time = time(&u->last_login);
  c->kb_up = c->kb_down = 0;
  server_log_local(c,LG_LOGIN,"LOGIN by %s (%s) from %s@%s (%s)",*(p+1),
    anonuser(u) ? *(p+2) : "********",c->ident,
    inet_ntoa(addr.sin_addr),hp ? hp->h_name : "unknown host");
}

void user_list_restricted(struct child_s* c, char** p, int n)
{ char           str[32];
  struct user_s* u;

  sprintf(str,"%i",user_count);
  server_reply(c,str);
  if (!fgets(str,sizeof(str),c->rdp)) return;
  if (*str != '+') return;
  u = user;
  for (n=0; n<user_count; n++)
  { sprintf(str,"%s %li",u->name,u->uid);
    server_reply(c,str);
    u++;
  }
}

void user_remove(struct child_s* c, char** p, int n)
{ struct child_s* ch = child;
  int             i;
  struct user_s*  u;

  if (!allowed(c->user,R_USER_REMOVE))
  { server_reply(c,"-You are not superuser.");
    return;
  }
  if (n != 1 || strlen(*p)>=sizeof(u->name))
  { server_reply(c,"-Syntax: USER REMOVE <name>");
    return;
  }
  u = user;
  for (n=0; n<user_count; n++, u++)
  { if (!strcasecmp(*p,u->name))
    { if (groupop(c->user) && !groupopof(c->user,u))
      { server_reply(c,"-User is not in your group.");
        return;
      }
      top_ok = 0;
      for (i=0; i<children; i++, ch++) if (ch->user)
      { if (ch->user == u)
        { kill(ch->pid,SIGKILL);
          child_logout(ch,CL_KICK);
        } else if (ch->user > u) ch->user--;
      }
      memmove(u,u+1,((--user_count)-n)*sizeof(struct user_s));
      server_reply(c,"+User removed.");
      return;
    }
  }
  server_reply(c,"-User does not exist.");
}

void user_reset(struct child_s* c, char** p, int n)
{ int            td;
  int            se = 0;
  struct user_s* u = user;

  if (!allowed(c->user,R_USER_CHANGE_STATS))
  { server_reply(c,"-You are not superuser.");
    return;
  }

  if (n != 1) se = 1;
  if (!se) if (!strcasecmp(*p,"up")) td = 1; else
    if (!strcasecmp(*p,"down")) td = 2; else
    if (!strcasecmp(*p,"credit")) td = 3; else se = 1;
  if (!se) for (n=0; n<user_count; n++, u++) if (!templateuser(u)) switch (td)
  { case 1 :
      u->kb_uploaded = 0;
      break;
    case 2 :
      u->kb_downloaded = 0;
      break;
    case 3 :
      if ( u->limit != -1 ) /* limit == -1 -> don't change credits anymore */
        u->credit = 0;
      break;
  }
  if (se)
  { server_reply(c,"-Syntax: RESET { up | down | credit }");
    return;
  }
  server_reply(c,"+Reset successful.");
}

long new_usercredit(struct child_s* c, long kb_up, long kb_down)
{ if (!c->user)
  { fprintf(stderr,"ftpd: c->user in new_usercredit() is NULL, c->pid=%i\n",c->pid);
    return 0;
  }
  if (anonuser(c->user))
  { c->kb_up = kb_up;
    c->kb_down = kb_down;
    if (c->user->limit == -1) /* limit == -1 -> don't change credits anymore */
      return c->user->credit;
    else if (c->user->limit)
      return c->user->credit - c->kb_down + c->kb_up * 100 / c->user->limit;
      else return 0;
  } else
  { c->user->kb_up_tmp += kb_up - c->kb_up;
    c->user->kb_down_tmp += kb_down - c->kb_down;
    c->kb_up = kb_up;
    c->kb_down = kb_down;
    if (c->user->limit == -1) /* limit == -1 -> don't change credits anymore */
      return c->user->credit;
    else if (c->user->limit)
      return c->user->credit - c->user->kb_down_tmp +
        c->user->kb_up_tmp * 100 / c->user->limit;
    else return 0;
  }
}

void user_upload_download(struct child_s* c, char** p, int n)
{ int  exitcode;
  long kb_down;
  long kb_up;

  if (n == 1)
  { if (!sscanf(*p,"%i",&exitcode))
    { server_reply(c,"\n");
      return;
    }
    server_reply(c,"\n");
    child_logout(c,exitcode);
  } else if (n == 2)
  { sscanf(*p,"%li",&kb_up);
    sscanf(*(p+1),"%li",&kb_down);
    sprintf(*p,"%li",new_usercredit(c,kb_up,kb_down));
    server_reply(c,*p);
  } else server_reply(c,"\n");
}

void user_upload_reduce(struct child_s* c, char** p, int n)
{ long           size;
  struct user_s* u;

  if ((n != 2) || !sscanf(*(p+1),"%li",&size))
  { server_reply(c,"-Syntax: UUR <name> <size>");
    return;
  }
  u = get_user_by_name(*p);
  if (!u)
  { server_reply(c,"-User does not exist.");
    return;
  }
  if (!anonuser(u))
  { top_ok = 0;
    size /= 1024;
    u->kb_uploaded -= size;
    if (u->limit && u->limit != -1) /* limit == -1 -> don't change credits anymore */
      u->credit -= size * 100 / u->limit;
  }
  server_reply(c,"+");
}

void server_reply_who(struct child_s* c, char *line, struct child_s* u, int n)
{ char cline[1024];
  char fline[256];
  char host[100];
  time_t t;
  
  *fline='\0';
  *cline='\0';
  while(*line)
  { if ((*line == '%') && (*(line+1)) && (*(line+1) == 'W') && (*(line+2)))
    { if ((*(line+2))=='[')
      { line++; line++; line++;
        while((*line) && (*line!=']'))
	{ *(fline+strlen(fline)+1) = '\0';
	  *(fline+strlen(fline)) = *line;
	  line++;
	}
	if ((*line==']') && (*(line+1)))
	{ line++;
	  switch (*line)
	  {
	    /***** %W[intformat]C ******* Counter **************************************/
	    case 'C' : sprintf(cline+strlen(cline),fline,n); break;

	    /***** %W[stringformat]N **** Username *************************************/
	    case 'N' : sprintf(cline+strlen(cline),fline,u->user->name); break;

	    /***** %W[stringformat]G **** Groupname ************************************/
	    case 'G' : sprintf(cline+strlen(cline),fline,get_group_by_gid(u->user->gid)); break;

	    /***** %W[timeformat]T ****** OnlineTime ***********************************/
	    case 'T' : t=time(NULL)-u->time; strftime(cline+strlen(cline),1023-strlen(cline),fline,gmtime(&t)); break;

	    /***** %W[timeformat]t ****** ActivityTime *********************************/
	    case 't' : t=time(NULL)-u->act_time; strftime(cline+strlen(cline),1023-strlen(cline),fline,gmtime(&t)); break;

	    /***** %W[longintformat]M *** Uploaded MB **********************************/
	    case 'M' : sprintf(cline+strlen(cline),fline,u->kb_up==0 ? u->kb_up : u->kb_up/1024 ); break;

	    /***** %W[longintformat]m *** Downloaded MB ********************************/
	    case 'm' : sprintf(cline+strlen(cline),fline,u->kb_down==0 ? u->kb_down : u->kb_down/1024 ); break;

	    /***** %W[stringformat]A **** Activity *************************************/
	    case 'A' : sprintf(cline+strlen(cline),fline,u->activity); break;

            /***** %W[stringformat]l **** Idletag  *************************************/
            case 'l' : sprintf(cline+strlen(cline),u->idle ? fline : " "); break;
     
	    /***** %W[stringformat]L **** InfoLine *************************************/
	    case 'L' : sprintf(cline+strlen(cline),fline,u->user->info); break;

	    /***** %W[stringformat]I **** Ident ****************************************/
	    case 'I' : sprintf(cline+strlen(cline),fline,u->ident); break;

	    /***** %W[stringformat]i **** IP *******************************************/
	    case 'i' : get_hostname(u->ip,host); sprintf(cline+strlen(cline),fline,host); break;

	    /***** %W[stringformat]o **** Ownlogin *********************************/
            case 'o' : sprintf(cline+strlen(cline),u->pid == c->pid ? fline : " "); break;

	    /***** %W[stringformat]H **** Ident@IP *************************************/
	    case 'H' :
	      strncpy(host,u->ident,strlen(u->ident));
	      host[strlen(u->ident)]='@'; 
	      host[strlen(u->ident)+1]='\0'; 
	      get_hostname(u->ip,host+strlen(host)); 
	      sprintf(cline+strlen(cline),fline,host);
	      break;

	    default :
	      strncpy(cline+strlen(cline),line-strlen(fline)-3,strlen(fline)+3);
	      *(cline+strlen(cline)+strlen(fline)+4) = '\0';
	      break;
	  }
	  *fline='\0';
	  line++;
	}
      }
    } else
    { *(cline+strlen(cline)+1) = '\0';
      *(cline+strlen(cline)) = *line;
      line++;
    }
  }
  server_reply(c,cline);
}

void user_who(struct child_s* c, char** p, int n)
{ int             i;
  int             k=0;
  struct child_s* u;
  char*           p1;
  char*           p2;
  char            line[1024];
  prepare_msg(&whoonline_msg_top);
  prepare_msg(&whoonline_msg_body);
  prepare_msg(&whoonline_msg_tail);
  /**********************************************/
  /* Cooking Who_Top! (Nothing to be cooked???) */
  /**********************************************/
  p1 = whoonline_msg_top.msg ? whoonline_msg_top.msg : whoonline_msg_top.default_msg;
  p2 = p1;
  do
  { while (*p2 && (*p2 != '\n')) p2++;
    if (*p2) p2++;
    line[0]='+';
    memcpy(line+1,p1,p2-p1);
    line[p2-p1+1]='\0';
    server_reply_who(c,line,c,0);
    p1 = p2;
  } while (*p2);
  /*********************/
  /* Cooking Who_Body! */
  /*********************/
  u = child;
  for (i=0; i<children; i++, u++) if (u->user && !invisible(u->user))
  { p1 = whoonline_msg_body.msg ? whoonline_msg_body.msg : whoonline_msg_body.default_msg;
    p2 = p1;
    do
    { while (*p2 && (*p2 != '\n')) p2++;
      if (*p2) p2++;
      line[0]='+';
      memcpy(line+1,p1,p2-p1);
      line[p2-p1+1]='\0';
      server_reply_who(c,line,u,k+1);
      p1 = p2;
    } while (*p2);
    k++;
  }
  /***********************************************/
  /* Cooking Who_Tail! (Nothing to be cooked???) */
  /***********************************************/
  p1 = whoonline_msg_tail.msg ? whoonline_msg_tail.msg : whoonline_msg_tail.default_msg;
  p2 = p1;
  do
  { while (*p2 && (*p2 != '\n')) p2++;
    if (*p2) p2++;
    line[0]='+';
    memcpy(line+1,p1,p2-p1);
    line[p2-p1+1]='\0';
    server_reply_who(c,line,c,k);
    p1 = p2;
  } while (*p2);
  server_reply(c,"+");
}

struct serverd_s
{ char command[4];
  void (*function)(struct child_s*,char**,int);
} serverd_cmd[] =
{ { "ADI", account_disable               },
  { "ADU", add_user                      },
  { "AEN", account_enable                },
  { "BIN", bind_port                     },
  { "CNT", count                         }, 
  { "COL", color                         },
  { "DGF", dump_group_file               },
  { "DUF", dump_user_file                },
  { "DUP", send_dupechk                  },
  { "FCM", file_chmod                    },
  { "FCO", file_chown                    },
  { "FDN", file_downloaded               },
  { "FRM", file_removed                  },
  { "FRN", file_renamed                  },
  { "FUP", file_uploaded                 },
  { "GAD", group_add                     },
  { "GCH", group_change                  },
  { "GLI", group_list                    },
  { "GLR", group_list_restricted         },
  { "GRM", group_remove                  },
  { "GRP", groups                        },
  { "KCK", kick_user                     },
  { "LOG", server_log                    },
  { "MAX", server_max                    },
  { "MCD", send_chdir_msg                },
  { "MCU", send_cdup_msg                 },
  { "MDE", send_dele_msg                 },
  { "MDN", send_download_msg             },
  { "MGB", send_goodbye_msg              },
  { "MLS", send_list_msg                 },
  { "MMD", send_mkdir_msg                },
  { "MSU", send_startup_msg              },
  { "MTA", send_toomanyanon_msg          },
  { "MTU", send_toomanyuser_msg          },
  { "MUP", send_upload_msg               },
  { "MWC", send_welcome_msg              },
  { "REC", server_reconfig               },
  { "SCC", set_cdpath                    },
  { "SCD", send_cdpath                   },
  { "SCH", send_check                    },
  { "SCL", server_close                  },
  { "SDC", send_dupecheck                },
  { "SDS", send_dshortcuts               }, 
  { "SIN", send_startup_info             },
  { "SOP", server_open                   },
  { "SSC", send_shortcut                 },
  { "SSD", server_shutdown               },
  { "STO", send_top                      },
  { "UAC", user_activity                 },
  { "UAD", user_add_n                    },
  { "UAN", user_anonymous                },
  { "UAS", user_add_su                   },
  { "UCH", user_change                   },
  { "UCP", user_change_password          },
  { "UIN", user_information              },
  { "ULD", user_list_dump                },
  { "ULI", user_list                     },
  { "ULL", user_list_loggedin            },
  { "ULP", user_request_login_permission },
  { "ULR", user_list_restricted          },
  { "URM", user_remove                   },
  { "URS", user_reset                    },
  { "UUD", user_upload_download          },
  { "UUR", user_upload_reduce            },
  { "VER", server_version                },
  { "WHO", user_who                      } };

void child_communication(struct child_s* c)
{ char  cmd[1024];
  char* p[20];
  int   i = 0;
  int   j;
  int   k = (sizeof(serverd_cmd) / sizeof(struct serverd_s)) - 1;
  int   n;
  int   r;

  switch (read_command(c,cmd,sizeof(cmd),p,16,&n))
  { case -1 :
      fprintf(stderr,"ftpd: pipe error, errno=%i, child_pid=%i, username=%s,"
        " logged out\n",errno,c->pid,c->user ? c->user->name : "???");
      kill(c->pid,SIGKILL);
      child_logout(c,CL_PIPEERR);
      return;
    case  1 :
      fprintf(stderr,"ftpd: pipe error (1)\n");
      server_reply(c,"-Too many parameters.");
      return;
    case  2 :
      fprintf(stderr,"ftpd: pipe error (2)\n");
      server_reply(c,"-Command string of zero length received.");
      return;
    case  3 :
      fprintf(stderr,"ftpd: pipe error (3)\n");
      server_reply(c,"-Command string too long.");
      return;
    case  4 :
      fprintf(stderr,"ftpd: pipe error (4)\n");
      server_reply(c,"-No command in command string.");
      return;
  }
#ifdef DEBUG
  fprintf(debugf,"child command",*p);
  for (i = 0; i < n; i++) fprintf(debugf," %s",*(p+i));
  fprintf(debugf,"\n");
  i = 0;
#endif
  while (i <= k)
  { j = (i + k) >> 1;
    r = strcasecmp(*p,serverd_cmd[j].command);
    if (r > 0)  i = j + 1; else if (r < 0) k = j - 1; else
    {
      serverd_cmd[j].function(c,p+1,n-1);
      return;
    }
  }
  fprintf(stderr,"ftpd(%i): unknown client command (%s)\n",getpid(),*p);
  server_reply(c,"-Unknown command.");
}


/* EOF */
