/*   transfer.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996,1997,1998,1999 by Christoph Schwarz
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <setjmp.h>
#include "tweak.h"
#ifdef HAVE_SYS_SELECT_H
  #include <sys/select.h>
#endif
#ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
#endif
#include "externs.h"
#include "perm.h"

jmp_buf scriptabortjmp;

void alarmhandler(int sig)
{ scriptabort = 1;
  longjmp(scriptabortjmp,1);
}

int cmd_mode(char** p, int n)
{ if (n != 1) return send_reply(501,"Syntax: MODE <mode_id>");
  if (!strcasecmp(*p,"S")) return send_reply(200,"Mode set to STREAM.");
  if (!strcasecmp(*p,"B")) return send_reply(504,"Block mode not supported.");
  if (!strcasecmp(*p,"C")) return send_reply(504,"Compressed mode not supported.");
  return send_reply(501,"Invalid mode identifier.");
}

int cmd_stru(char** p, int n)
{ if (n != 1) return send_reply(501,"Syntax: STRU <structure_id>");
  if (!strcasecmp(*p,"F"))
  { structure = S_FILE;
    return send_reply(200,"Structure set to FILE.");
  }
  if (!strcasecmp(*p,"R"))
  { structure = S_RECORD;
    return send_reply(200,"Structure set to RECORD.");
  }
  if (!strcasecmp(*p,"P"))
    return send_reply(504,"Page structure not supported.");
  return send_reply(501,"Invalid structure identifier.");
}

int cmd_type(char** p, int n)
{ if (n<1 || n>2) return send_reply(501,"Syntax: TYPE <type_id>");
  if (!strcasecmp(*p,"A"))
  { type = T_ASCII;
    return send_reply(200,"Type set to ASCII.");
  }
  if (!strcasecmp(*p,"I"))
  { type = T_IMAGE;
    return send_reply(200,"Type set to IMAGE.");
  }
  if (!strcasecmp(*p,"E")) return send_reply(504,"EBCDI type not supported.");
  if (!strcasecmp(*p,"L")) return send_reply(504,"LOCAL type not supported.");
  return send_reply(501,"Invalid type identifier.");
}

int cmd_mdtm(char** p, int n)
{ char        dir[256];
  struct stat st;

  if (n != 1) return send_reply(501,"Syntax: MDTM <filename>");
  if ((add_dir(cwd,*p,dir,sizeof(dir)) == -1) || (stat(dir+1,&st) == -1))
    return send_error(550,"mdtm",0);
  if (st.st_mode & S_IFREG)
  { strftime(dir,sizeof(dir),"%Y%m%d%H%M%S",localtime(&st.st_mtime));
    return send_reply(213,dir);
  } else send_message(550,1,"%s: not a plain file.",*p);
}

int cmd_size(char** p, int n)
{ char        dir[256];
  struct stat st;

  if (n != 1) return send_reply(501,"Syntax: SIZE <filename>");
  if ((add_dir(cwd,*p,dir,sizeof(dir)) == -1) || (stat(dir+1,&st) == -1))
    return send_error(550,"size",0);
  if (st.st_mode & S_IFREG)
  { sprintf(dir,"%li",st.st_size);
    return send_reply(213,dir);
  } else send_message(550,1,"%s: not a plain file.",*p);
}

int cmd_rest(char** p, int n)
{ int se = 0;

  if (type == T_ASCII)
    return send_reply(502,"REST not possible with ASCII type.");
  if (n == 1)
  { if (!sscanf(*p,"%li",&rest_offset)) se = -1;
  } else se = -1;
  if (se == -1) return send_reply(501,"Syntax: REST <offset>");
  return send_reply(350,"REST command successful.");
}

int limit_ok(long size)
{ if (superuser() || !usr.limit) return 1;
  return ((size / 1024) <= usr.credit);
}

cmd_transferpermitted(char* prog,char* buffer)
{ char  str[255];
  int   ec;
  char* p[8];

  p[0] = prog;
  p[1] = buffer;

  if (run_prog_xt(prog,p,&ec,&str) == -1) return 0;

  if (WIFEXITED(ec))
  { if (WEXITSTATUS(ec) && (WEXITSTATUS(ec) != -1)) ec = 550; else ec = 150;
  } else ec = 150;
  if (ec == 550)
  { sprintf(buffer,"Transfer initial check failed (%s).",str);
    if (*buffer) if (send_message(ec,ec == 550,buffer) == -1) return -1;
  }
  return ec == 550 ? 1 : 0;
}

int cmd_retr(char** p, int n)
{ 
  char           buffer[4096];
  char*          dir;
  long           down = 0;
  int            ds;
  fd_set         fdset_r;
  fd_set         fdset_w;
  FILE*          file;
  int            len;
  int            limited;
  char*          name;
  FILE*          sock;
  struct stat    st;
  struct timeval t;
  time_t         t1;
  time_t         t2;
  struct sockaddr_in srv_addr;

  if (n != 1) return send_reply(501,"Syntax: RETR <filename>");
  ds = add_dir(cwd,*p,buffer,sizeof(buffer));
  split_dir(*p,&dir,&name,0);

  if (*pre_download) switch (cmd_transferpermitted(pre_download,buffer))
                       { case -1 : return -1;
                         case  1 : return 0;
                       }

  client_activity("DOWNLOAD %s",name);
  client_log_start(LG_GET,"DOWNLOAD %s",ds == -1 ? name : buffer);
  if (type == T_ASCII) file = file_open(dir,name,"rt",P_READ);
    else file = file_open(dir,name,"rb",P_READ);
  if (!file) return send_error(550,"read",LG_GET);
  if ((type == T_IMAGE) && rest_offset)
  { if (fseek(file,rest_offset,SEEK_SET) == -1)
       return send_error(550,"seek",LG_GET);
    rest_offset = 0;
  }
  limited = permission_part_ok(dir,P_NOLIMIT,1);
  if (fstat(fileno(file),&st) == -1) st.st_size = 0;
  if (!(st.st_mode & S_IFREG))
  { fclose(file);
    client_log_finish(LG_GET," FAILED: not a plain file");
    return send_message(550,1,"%s: not a plain file.",name);
  }
  if (limited && !limit_ok(st.st_size))
  { fclose(file);
    sprintf(buffer,"Not enough credit (you have: %li KB / you need: %li KB)",
      usr.credit,st.st_size / 1024);
    client_log_finish(LG_GET," FAILED: %s",buffer);
    return send_reply(550,buffer);
  }
  if (pasvsock == -1)
  { if ((ds = socket(PF_INET,SOCK_STREAM,0)) == -1)
    { fclose(file);
      return send_error(425,"socket",LG_GET);
    }
    if (other_ip_s != 0)
    { srv_addr.sin_family = AF_INET;
      srv_addr.sin_port = 0;
      srv_addr.sin_addr.s_addr = other_ip_s;
      bzero(srv_addr.sin_zero, sizeof srv_addr.sin_zero);
      if (bind(ds,(struct sockaddr*)&srv_addr,sizeof(srv_addr)) == -1)
      { fclose(file);
        return send_error(425,"bind",LG_GET);
      }
    }
    if (connect(ds,(struct sockaddr*)&addr,sizeof(addr)) == -1)
    { fclose(file);
      return send_error(425,"connect",LG_GET);
    }
  } else
  { alarm(timeout);
    len = sizeof(addr);
    ds = accept(pasvsock,(struct sockaddr*)&addr,&len);
    alarm(0);
    if (ds == -1)
    { fclose(file);
      return send_error(425,"accept",LG_GET);
    }
  }
  sock = fdopen(ds,type == T_ASCII ? "wt" : "wb");
  if (!sock)
  { len = errno;
    close(ds);
    fclose(file);
    errno = len;
    return send_error(425,"fdopen",LG_GET);
  }
  sprintf(buffer,"Opening %s data connection for %s (%li bytes).",
    type == T_ASCII ? "ASCII" : "BINARY", name,st.st_size);
  if (send_reply(150,buffer) == -1)
  { client_log_error(LG_GET);
    fclose(file);
    fclose(sock);
    return -1;
  }
  time(&t1);
  FD_ZERO(&fdset_r);
  FD_ZERO(&fdset_w);
  while (n)
  { FD_SET(fileno(sockf),&fdset_r);
    FD_SET(ds,&fdset_w);
    t.tv_sec  = timeout;
    t.tv_usec = 0;
    n = select(FD_SETSIZE,&fdset_r,&fdset_w,NULL,&t);
    if (!n)
    { client_log_finish(LG_GET," FAILED: Timeout");
      fclose(file);
      fclose(sock);
      return send_reply(426,"Timeout, data connection closed");
    }
    if (n == -1)
    { fclose(file);
      fclose(sock);
      return send_error(426,"select",LG_GET);
    }
    if (FD_ISSET(fileno(sockf),&fdset_r))
    { switch (read_ftp_command(buffer,sizeof(buffer),p,4,&n))
      { case -1 :
          fclose(file);
          fclose(sock);
          client_log_error(LG_GET);
          return -1;
        case  0 :
          if (strcasecmp(*p,"ABOR"))
          { if (send_reply(503,"Unexpected command.") == -1)
            { fclose(file);
              fclose(sock);
              client_log_error(LG_GET);
              return -1;
            }
          } else
          { client_log_finish(LG_GET," FAILED: Transfer interrupted");
            fclose(file);
            fclose(sock);
            if (send_reply(426,"Transfer interrupted.") == -1) return -1;
            return send_reply(226,"Transfer successfully aborted.");
          }
      }
    } else
    { if (type == T_ASCII)
      { if (!fgets(buffer,sizeof(buffer)-2,file)) *buffer = '\0';
        n = strlen(buffer);
        if (((n == 1) && (*buffer == '\n')) || ((n >= 2) &&
          (*(buffer+n-1) == '\n') && (*(buffer+n-1) != '\r')))
        { *(buffer+n-1) = '\r';
          *(buffer+n) = '\n';
          n++;
        }
        *(buffer+n) = '\0';
        if (n) n = fputs(buffer,sock);
      } else
      { n = fread(buffer,1,sizeof(buffer),file);
        if (n) n = fwrite(buffer,1,n,sock);
      }
      if (n == -1)
      { n = send_error(426,"send",LG_GET);
        fclose(file);
        fclose(sock);
        return n;
      }
      down += n;
    }
  }
  fstat(fileno(file),&st);
  fclose(file);
  fclose(sock);
  time(&t2);
  t2 -= t1;
  if (!t2) t2 = 1;
  t2 = down / t2;
  last_speed = t2;
  if (t2 >= 1024)
    client_log_finish(LG_GET," (%li bytes @ %3.1f Kbytes/sec)",down,(float)t2/1024);
    else client_log_finish(LG_GET," (%li bytes @ %li bytes/sec)",down,t2);
  if (st.st_size) down = st.st_size;
  if (limited)
  { mbytes_downloaded += down / 1048576;
    bytes_downloaded = (down+bytes_downloaded) % 1048576;
    printf("UUD %li %li\n",mbytes_uploaded*1024 + bytes_uploaded/1024,mbytes_downloaded*1024 + bytes_downloaded/1024);
    scanf("%li",&usr.credit);
    getc(stdin);
  }
  *(name-1) = '/';
  add_dir(cwd,*p,buffer,sizeof(buffer));
  printf("FDN- %s\n",buffer);
  fgets(buffer,sizeof(buffer),stdin);
  return send_message(226,1,download_msg,name);
}

int cmd_stor_check(char* buffer, char* dir, char* name)
{ struct check_s* ch = check;
  int             i;
  char*           p[8];
  void            (*oldvec)();

  for (i=0; i<check_count; i++, ch++) if (!strcasecmp(name+strlen(name)-
    strlen(ch->extension),ch->extension))
  { if (add_dir(cwd,dir,buffer,512) == -1) return 0;
    i = 0;
    p[i++] = ch->name;
    if (*ch->param) p[i++] = ch->param;
    p[i++] = name;
    p[i++] = usr.name;
    if (checktimeout)
    { if (setjmp(scriptabortjmp))
      { if (scriptabort == 1) { scriptabort = 0; return 2; } }
      oldvec = signal(SIGALRM, alarmhandler);
      alarm(checktimeout);
    }
    if (run_program(ch->name,p,i,0,0,buffer+1,&i,buffer) == -1) return 0;
    if (checktimeout)
    { alarm(0);
      signal(SIGALRM, oldvec);
    }
    if (WIFEXITED(i))
    { if (WEXITSTATUS(i) && (WEXITSTATUS(i) != -1)) i = 550; else i = 226;
    } else i = 226;
    if ((i == 550) && !*buffer) strcpy(buffer,"File is corrupt.");
    if (*buffer) if (send_message(i,i == 550,buffer) == -1) return -1;
    if (i == 550)
    { add_dir(cwd,dir,buffer,512);
      add_dir(buffer,name,buffer,512);
      unlink(buffer+1);
    }
    return i == 550 ? 1 : 0;
  }
  return 0;
}

int cmd_stor_dcheck(char* buffer, char* name)
{ struct check_s* ch = dcheck;
  int             i;
  char*           p[8];

  for (i=0; i<dcheck_count; i++, ch++) if (!strcasecmp(name+strlen(name)-
    strlen(ch->extension),ch->extension))
  { i = 0;
    p[i++] = ch->name;
    if (*ch->param) p[i++] = ch->param;
    p[i++] = buffer + 1;
    p[i++] = usr.name;
    if (run_program(ch->name,p,i,0,0,NULL,&i,buffer) == -1) return 0;
    if (WIFEXITED(i))
    { if (WEXITSTATUS(i) && (WEXITSTATUS(i) != -1)) i = 550; else i = 150;
    } else i = 150;
    if ((i == 550) && !*buffer) strcpy(buffer,"Dupe check failed.");
    if (*buffer) if (send_message(i,i == 550,buffer) == -1) return -1;
    return i == 550 ? 1 : 0;
  }
  return 0;
}

void cmd_stor_rm(char* fullname, char* dir, char* name)
{ if (add_dir(cwd,dir,fullname,512) == -1) return;
  if (add_dir(fullname,name,fullname,512,0) == -1) return;
  unlink(fullname+1);
}

int cmd_stor(char** p, int n)
{ char           buffer[4096];
  char*          dir;
  int            ds;
  FILE*          dsf;
  fd_set         fdset_r;
  FILE*          file;
  int            limited;
  char*          name;
  int            ok;
  struct stat    st;
  struct timeval t;
  time_t         t1;
  time_t         t2;
  long           up = 0;

  if (n != 1) return send_reply(501,"Syntax: STOR <filename>");
  convert_filename(*p,fconv_mode & 0x0F);
  ok = add_dir(cwd,*p,buffer,sizeof(buffer));
  if (ok != -1) if (stat(buffer+1,&st) == -1) st.st_size = 0;
  split_dir(*p,&dir,&name,0);
  if (strlen(name) >= MAXFILENAMELENGTH)
  { errno = ENAMETOOLONG;
    return send_error(550,"open",0);
  }
  if ((permission_ok(dir,P_WRITE) == -1) || (!superuser() && (0==strcmp(name,readme_file))) || (0 == strcmp(name,permission_file)))
      return send_error(550,"write",0);

  if (*pre_upload) switch (cmd_transferpermitted(pre_upload,buffer))
                     { case -1 : return -1;
                       case  1 : return 0;
                     }
  switch (cmd_stor_dcheck(buffer,name))
  { case -1 : return -1;
    case  1 : return 0;
  }
  if (find_dupe(name,buffer))
    return send_message(550,1,"Dupe detected (%s)",buffer);
  file = file_open(dir,name,type == T_ASCII ? "wt" : "wb",P_WRITE);
  if (!file) return send_error(550,"write",0);
  client_activity("UPLOAD %s",name);
  *(name-1) = '/';
  ok = add_dir(cwd,*p,buffer,sizeof(buffer));
  printf("FCO- %li %li %s\n",usr.uid,-2,buffer);
  fgets(buffer+2048, sizeof(buffer)-2048, stdin);
  client_log_start(LG_PUT,"UPLOAD %s",ok == -1 ? *p : buffer);
  split_dir(*p,&dir,&name,0);
  limited = permission_part_ok(dir,P_NOLIMIT,0);
  if (pasvsock == -1)
  { if ((ds = socket(PF_INET,SOCK_STREAM,0)) == -1)
    { fclose(file);
      return send_error(425,"socket",LG_PUT);
    }
    if (connect(ds,(struct sockaddr*)&addr,sizeof(addr)) == -1)
    { n = errno;
      fclose(file);
      errno = n;
      return send_error(425,"connect",LG_PUT);
    }
  } else
  { alarm(timeout);
    ok = sizeof(addr);
    ds = accept(pasvsock,(struct sockaddr*)&addr,&ok);
    alarm(0);
    if (ds == -1)
    { n = errno;
      fclose(file);
      cmd_stor_rm(buffer,dir,name);
      errno = n;
      return send_error(425,"accept",LG_PUT);
    }
  }
  if (type == T_ASCII)
  { dsf = fdopen(ds,"rt");
    if (!dsf)
    { n = errno;
      close(ds);
      fclose(file);
      cmd_stor_rm(buffer,dir,name);
      errno = n;
      return send_error(425,"fdopen",LG_PUT);
    }
    sprintf(buffer,"Opening ASCII data connection for %s",name);
  } else sprintf(buffer,"Opening BINARY data connection for %s",name);
  if (send_reply(150,buffer) == -1)
  { client_log_error(LG_PUT);
    fclose(file);
    close(ds);
    cmd_stor_rm(buffer,dir,name);
    return -1;
  }
  time(&t1);
  FD_ZERO(&fdset_r);
  while (n)
  { FD_SET(fileno(sockf),&fdset_r);
    FD_SET(ds,&fdset_r);
    t.tv_sec  = timeout;
    t.tv_usec = 0;
    n = select(FD_SETSIZE,&fdset_r,NULL,NULL,&t);
    if (!n)
    { client_log_finish(LG_PUT," FAILED: Timeout");
      fclose(file);
      close(ds);
      cmd_stor_rm(buffer,dir,name);
      return send_reply(426,"Timeout, data connection closed\n");
    }
    if (n == -1)
    { n = errno;
      fclose(file);
      close(ds);
      cmd_stor_rm(buffer,dir,name);
      errno = n;
      return send_error(426,"select",LG_PUT);
    }
    if (FD_ISSET(fileno(sockf),&fdset_r))
    { switch (read_ftp_command(buffer,sizeof(buffer),p,4,&n))
      { case -1 :
          client_log_error(LG_PUT);
          return -1;
        case  0 :
          if (strcasecmp(*p,"ABOR"))
          { if (send_reply(503,"Unexpected command.") == -1)
            { client_log_error(LG_PUT);
              cmd_stor_rm(buffer,dir,name);
              return -1;
            }
          } else
          { client_log_finish(LG_PUT," FAILED: Transfer interrupted");
            fclose(file);
            if (type == T_ASCII) fclose(dsf); else close(ds);
            cmd_stor_rm(buffer,dir,name);
            if (send_reply(426,"Transfer interrupted.") == -1) return -1;
            return send_reply(226,"Transfer successfully aborted.");
          }
      }
    } else
    { if (type == T_ASCII)
      { if (!fgets(buffer,sizeof(buffer)-2,dsf)) *buffer = '\0';
        n = strlen(buffer);
        if ((n >= 2) && (*(buffer+n-2) == '\r'))
        { n--;
          *(buffer+n-1) = '\n';
          *(buffer+n) = '\0';
        }
        if (n) ok = fputs(buffer,file) != EOF; else ok = 1;
      } else
      { n = recv(ds,buffer,sizeof(buffer),0);
        if (n == -1)
        { fclose(file);
          close(ds);
          cmd_stor_rm(buffer,dir,name);
          return send_error(452,"recv",LG_PUT);
        }
        ok = (fwrite(buffer,1,n,file) == n);
      }
      if (!ok)
      { n = send_error(452,"write",LG_PUT);
        fclose(file);
        if (type == T_ASCII) fclose(dsf); else close(ds);
        cmd_stor_rm(buffer,dir,name);
        return n;
      }
      up += n;
    }
  }
  fclose(file);
  close(ds);
  time(&t2);
  t2 -= t1;
  if (!t2) t2 = 1;
  t2 = up / t2;
  if ((ok = cmd_stor_check(buffer,dir,name)) != 0)
  { if (ok == 2)
    { if (send_message(226,0,"Upload checker timeout. File not tested.",NULL) == -1) return -1;
      client_log(LG_PUT," CHECK FAILED: Upload script corrupt.");
    }
    else
    {  
      client_log_finish(LG_PUT," FAILED: File is corrupt");
      return ok;
    }
  }
  last_speed = t2;
  if (t2 >= 1024)
    client_log_finish(LG_PUT," (%li bytes @ %3.1f Kbytes/sec)",up,(float)t2/1024);
    else client_log_finish(LG_PUT," (%li bytes @ %li bytes/sec)",up,t2);
  if (limited)
  { mbytes_uploaded += (up-st.st_size) / 1048576; 
    bytes_uploaded = (up-st.st_size+bytes_uploaded) % 1048576;
    printf("UUD %li %li\n",mbytes_uploaded*1024 + bytes_uploaded/1024,mbytes_downloaded*1024 + bytes_downloaded/1024);
    scanf("%li",&usr.credit);
    getc(stdin);
  }
  *(name-1) = '/';
  add_dir(cwd,*p,buffer,sizeof(buffer));
  printf("FUP- %s\n",buffer);
  fgets(buffer,sizeof(buffer),stdin);
  return send_message(226,1,upload_msg,name);
}

int cmd_port(char** p, int n, char* cmd, int cmdlen)
{ int  a1;
  int  a2;
  int  a3;
  int  a4;
  int  p1;
  int  p2;
  char str[32];

  if (n != 1) return send_reply(501,"Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
  rest_offset = 0;
  if (sscanf(*p,"%i,%i,%i,%i,%i,%i",&a1,&a2,&a3,&a4,&p1,&p2) != 6)
    return send_reply(501,"Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
  if (a1<0 || a1>255 || a2<0 || a2>255 || a3<0 || a3>255 || a4<0 || a4>255 ||
      p1<0 || p1>255 || p2<0 || p2>255)
    return send_reply(501,"Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
  sprintf(str,"%i.%i.%i.%i",a1,a2,a3,a4);
  memset(&addr,0,sizeof(addr));
  addr.sin_addr.s_addr = inet_addr(str);
  if (addr.sin_addr.s_addr == -1) return send_reply(501,"Invalid address.");
  p1 = (p1<<8)+p2;
  addr.sin_port = htons((u_short)p1);
  addr.sin_family = AF_INET;
  if (pasvsock != -1) close(pasvsock);
  pasvsock = -1;
  return send_reply(200,"PORT command successful.");
}

int cmd_pasv(char** p, int n, char* cmd, int cmdlen)
{ struct hostent* hp;
  int             len;
  char            str[64];

  if (n) return send_reply(501,"Syntax: PASV");
  rest_offset = 0;
  if (pasvsock != -1) close(pasvsock);
  if (other_ip_s == 0) { gethostname(str,sizeof(str)-1);
                         hp = gethostbyname(str); 
                         if (!hp) return send_reply(550,"gethostbyname() failed"); }
     else              { hp=gethostbyaddr((char*)&other_ip_s,sizeof(other_ip_s),AF_INET);
                         if (!hp) return send_reply(550,"gethostbyaddr() failed"); }
  pasvsock = socket(PF_INET,SOCK_STREAM,0);
  if (pasvsock == -1) return send_error(550,"socket",0);
  addr.sin_family      = AF_INET;
  if (other_ip_s == 0) { addr.sin_addr.s_addr = htonl(INADDR_ANY); }
     else { addr.sin_addr.s_addr = other_ip_s; }
  addr.sin_port        = 0;
  if (bind(pasvsock,(struct sockaddr*)&addr,sizeof(addr)) == -1)
  { close(pasvsock);
    pasvsock = -1;
    return send_error(550,"bind",0);
  }
  if (listen(pasvsock,1) == -1)
  { close(pasvsock);
    pasvsock = -1;
    return send_error(550,"listen",0);
  }
  len = sizeof(addr);
  if (getsockname(pasvsock,(struct sockaddr*)&addr,&len) == -1)
  { close(pasvsock);
    pasvsock = -1;
    return send_error(550,"getsockname",0);
  }
  addr.sin_addr.s_addr = htonl((*(struct in_addr*)*hp->h_addr_list).s_addr);
  addr.sin_port = htons((u_short)addr.sin_port);
  sprintf(str,"%i,%i,%i,%i,%i,%i",(addr.sin_addr.s_addr>>24) & 0xFF,
   (addr.sin_addr.s_addr>>16) & 0xFF,(addr.sin_addr.s_addr>>8) & 0xFF,
   addr.sin_addr.s_addr & 0xFF,(addr.sin_port>>8) & 0xFF,addr.sin_port & 0xFF);
  if (send_message(227,1,"Entering Passive Mode (%s).",str) == -1)
  { close(pasvsock);
    pasvsock = -1;
    return -1;
  } 
  return 0;
}

/* EOF */
